import numpy as np
import pandas as pd
from Trackrefiner.core.correction.action.helper import extract_bacteria_features
from Trackrefiner.core.correction.action.featuresCalculation.calcElongationRate import calculate_elongation_rate
from Trackrefiner.core.correction.action.featuresCalculation.calcVelocity import calc_average_velocity
from Trackrefiner.core.correction.action.featuresCalculation.fluorescenceIntensity import determine_final_cell_type


def process_bacterial_life_and_family(processed_cp_out_df, interval_time, elongation_rate_method, assigning_cell_type,
                                      cell_type_array, label_col, center_coord_cols):
    """
    Calculates additional features related to the life history of bacteria from a processed
    CellProfiler output DataFrame.


    :param pandas.DataFrame processed_cp_out_df:
        Processed CellProfiler output DataFrame containing measured bacterial features.

    :param float interval_time:
        Time interval between consecutive images (in minutes).

    :param str elongation_rate_method:
        Method for calculating the Elongation rate. Options:
        - 'Average': Computes the average elongation rate.
        - 'Linear Regression': Estimates the elongation rate using linear regression.

    :param bool assigning_cell_type:
        If True, assigns cell types to objects based on intensity thresholds.

    :param numpy.ndarray cell_type_array:
        A boolean array indicating the channel in which the intensity exceeds the threshold for each bacterium.

    :param str label_col:
        Column name in the CellProfiler DataFrame that indicates the label of bacteria during tracking.

    :param dict center_coord_cols:
        A dictionary with keys 'x' and 'y', specifying the column names for center coordinates.

    :returns:
        processed_cp_out_df (pandas.DataFrame):
            A DataFrame including new features related to the life history and family of bacteria.
        data_frame_with_selected_col (pandas.DataFrame):
            A subset of `processed_cp_out_df` that includes specific columns to save in numpy output files.


    """

    processed_cp_out_df["time"] = pd.to_numeric(processed_cp_out_df["ImageNumber"]) * interval_time
    processed_cp_out_df["radius"] = processed_cp_out_df['AreaShape_MinorAxisLength'] / 2
    processed_cp_out_df['Average_Length'] = \
        processed_cp_out_df.groupby('id')['AreaShape_MajorAxisLength'].transform('mean')

    columns_to_initialize = ["Elongation_Rate", "startVol", "targetVol", "pos", "dir", "ends",
                             "strainRate", "strainRate_rolling", 'Velocity', 'Instant_Elongation_Rate',
                             'Instant_Velocity', 'Division_Family_Count']

    for column in columns_to_initialize:
        processed_cp_out_df[column] = ''

    # useful for calculation of rolling average
    window_size = 5  # time steps

    bacteria_id = processed_cp_out_df['id'].unique()

    for bacterium_id in bacteria_id:

        bacterium_life_history = processed_cp_out_df.loc[processed_cp_out_df['id'] == bacterium_id]
        elongation_rate = calculate_elongation_rate(bacterium_life_history, interval_time, elongation_rate_method)
        velocity = calc_average_velocity(bacterium_life_history, center_coord_cols, interval_time)

        # length of bacteria when they are born
        birth_length = bacterium_life_history["AreaShape_MajorAxisLength"].values[0]
        division_length = bacterium_life_history["AreaShape_MajorAxisLength"].values[-1]

        strain_rate_list = []
        old_length = birth_length
        old_position = np.sqrt(bacterium_life_history[center_coord_cols['x']].values[0] ** 2 +
                               bacterium_life_history[center_coord_cols['y']].values[0] ** 2)

        processed_cp_out_df.loc[bacterium_life_history.index, "Elongation_Rate"] = elongation_rate
        processed_cp_out_df.loc[bacterium_life_history.index, "Velocity"] = velocity
        processed_cp_out_df.loc[bacterium_life_history.index, "startVol"] = birth_length
        processed_cp_out_df.loc[bacterium_life_history.index, "targetVol"] = division_length

        for idx, bacterium in bacterium_life_history.iterrows():

            # https://github.com/cellmodeller/CellModeller/blob/master/CellModeller/Biophysics/BacterialModels/CLBacterium.py#L674
            strain_rate = (bacterium["AreaShape_MajorAxisLength"] - old_length) / old_length
            processed_cp_out_df.at[idx, "strainRate"] = strain_rate

            # rolling average
            strain_rate_list.append(strain_rate)
            if len(strain_rate_list) > window_size:
                # ignore first element
                strain_rate_rolling = np.mean(strain_rate_list[1:])
            else:
                strain_rate_rolling = np.mean(strain_rate_list)

            if bacterium['age'] > 1:
                instantaneous_elongation_rate = (
                    round((bacterium["AreaShape_MajorAxisLength"] - old_length) / interval_time, 3))

                # instantaneous velocity
                pos2 = np.sqrt(bacterium[center_coord_cols['x']] ** 2 + bacterium[center_coord_cols['y']] ** 2)
                instantaneous_velocity = round((pos2 - old_position) / interval_time, 3)

            else:
                instantaneous_elongation_rate = np.nan
                instantaneous_velocity = np.nan

            old_length = bacterium["AreaShape_MajorAxisLength"]
            old_position = np.sqrt(bacterium[center_coord_cols['x']] ** 2 + bacterium[center_coord_cols['y']] ** 2)

            processed_cp_out_df.at[idx, "strainRate_rolling"] = strain_rate_rolling
            processed_cp_out_df.at[idx, 'Instant_Elongation_Rate'] = instantaneous_elongation_rate
            processed_cp_out_df.at[idx, 'Instant_Velocity'] = instantaneous_velocity

            bacterium_features = extract_bacteria_features(bacterium, center_coord_cols)

            bacterium_center_position = [bacterium_features['center_x'], bacterium_features['center_y']]
            processed_cp_out_df.at[idx, "pos"] = bacterium_center_position

            processed_cp_out_df.at[idx, "dir"] = [np.cos(bacterium["AreaShape_Orientation"]),
                                                  np.sin(bacterium["AreaShape_Orientation"])]

            # find end points
            end_points = [[bacterium['Endpoint1_X'], bacterium['Endpoint1_Y']],
                          [bacterium['Endpoint2_X'], bacterium['Endpoint2_Y']]]

            processed_cp_out_df.at[idx, "ends"] = end_points

    processed_cp_out_df['Average_Instant_Velocity'] = \
        processed_cp_out_df.groupby('id')['Instant_Velocity'].transform('mean')

    if assigning_cell_type:
        # determine final cell type of each bacterium
        processed_cp_out_df = determine_final_cell_type(processed_cp_out_df, cell_type_array)

    # Iterate over each unique label.
    for label in processed_cp_out_df[label_col].unique():
        # Filter the DataFrame for rows corresponding to the current label.
        df_current_label = processed_cp_out_df.loc[processed_cp_out_df[label_col] == label]

        # Find rows where division occurred.
        mothers_df = df_current_label.loc[~ df_current_label["Max_Daughter_Mother_Length_Ratio"].isna()]

        processed_cp_out_df.loc[df_current_label.index, 'Division_Family_Count'] = mothers_df.shape[0]

    # rename some columns
    processed_cp_out_df = processed_cp_out_df.rename(columns={'age': 'cellAge'})

    data_frame_with_selected_col = \
        processed_cp_out_df.rename(columns={'ImageNumber': 'stepNum', 'AreaShape_MajorAxisLength': 'length',
                                            label_col: 'label', 'age': 'cellAge', 'Elongation_Rate': 'growthRate'})
    if assigning_cell_type:
        data_frame_with_selected_col = data_frame_with_selected_col[
            ['stepNum', 'ObjectNumber', 'id', 'label', 'divideFlag', 'cellAge', 'growthRate', 'LifeHistory', 'startVol',
             'targetVol', 'parent_id', 'pos', 'time', 'radius', 'length', 'ends', 'dir', 'cellType', 'strainRate',
             'strainRate_rolling']]
    else:
        data_frame_with_selected_col = data_frame_with_selected_col[
            ['stepNum', 'ObjectNumber', 'id', 'label', 'divideFlag', 'cellAge', 'growthRate', 'LifeHistory', 'startVol',
             'targetVol', 'parent_id', 'pos', 'time', 'radius', 'length', 'ends', 'dir', 'strainRate',
             'strainRate_rolling']]

    return processed_cp_out_df, data_frame_with_selected_col



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import os


def calculate_frequency(array, feature_related_col_name):
    """
    Calculate the frequency of unique values in a given array and return the result as a DataFrame.

    This function computes the frequency of each unique value in the input array and returns
    the results in a pandas DataFrame with columns for the unique values and their respective frequencies.

    :param numpy.ndarray array:
        A 1D array of values for which the frequency distribution is to be calculated.
    :param str feature_related_col_name:
        The name to use for the column representing unique values in the output DataFrame.

    :returns:
        pandas.DataFrame: A DataFrame with the following columns:
        - '<feature_related_col_name>': The unique values from the array.
        - 'Frequency': The count of each unique value.
    """

    # Get unique values and their counts
    unique_values, counts = np.unique(array, return_counts=True)

    # Create a DataFrame
    df = pd.DataFrame({
        feature_related_col_name: unique_values,
        'Frequency': counts
    })

    return df


def plot_frequency_distribution(x_positions, grouped_sel_feature, bin_labels, feature, feature_name, y_axis_value,
                                output_path):
    """
    Plot the frequency distribution of a selected feature and save the output.

    :param numpy.ndarray x_positions:
        An array of x-axis positions for the grouped frequency bars.
    :param pandas.DataFrame grouped_sel_feature:
        A DataFrame containing grouped frequency data to plot.
    :param list bin_labels:
        Labels for the x-axis bins.
    :param str feature:
        The name of the feature being plotted.
    :param str feature_name:
        A descriptive label for the feature (used for x-axis labeling).
    :param str y_axis_value:
        A descriptive label for y-axis labeling.
    :param str output_path:
        The directory path where the plot will be saved.
    """

    # Plot configuration
    fig, ax = plt.subplots(figsize=(10, 6))

    # Plot bars for cp_omnipose_abs_error and tr_abs_error
    ax.bar(x_positions, grouped_sel_feature["Frequency"], label="Trackrefiner", color="blue")

    # Add labels and title
    ax.set_xlabel(feature_name, fontsize=16, fontweight='bold', labelpad=15)
    ax.set_ylabel(y_axis_value, fontsize=16, fontweight='bold', labelpad=15)
    ax.set_title('')
    ax.set_xticks(x_positions)
    ax.set_xticklabels(bin_labels, rotation=90)
    plt.xticks(fontsize=14)
    plt.yticks(fontsize=14)

    # Show the plot
    plt.tight_layout()
    plt.savefig(f'{output_path}/plots/{feature}.jpg', dpi=600)
    plt.clf()
    plt.close()


def draw_feature_distribution(df, features_dict, label_col, interval_time, doubling_time, output_path):
    """
    Generate and plot the distribution of selected features from a DataFrame.

    This function iterates through a dictionary of features, calculates their frequency distribution,
    assigns them to bins, and visualizes the distribution with bar plots.

    :param pandas.DataFrame df:
        The input DataFrame containing feature data to analyze.
    :param dict features_dict:
        A dictionary mapping feature column names to descriptive labels and type of analysis for plotting.
    :param str label_col:
        Name of the column for object labels.
    :param float interval_time:
        Time interval to scale the 'LifeHistory' feature.
    :param float doubling_time:
        Threshold for filtering 'LifeHistory' values.
    :param str output_path:
        The directory path where the plots will be saved.

    :returns:
        None. The plots are saved as JPG files in the specified output directory.
    """

    os.makedirs(f'{output_path}/plots/', exist_ok=True)

    for feature in features_dict.items():

        feature_related_col_name = feature[0]
        feature_name = feature[1][0]
        y_axis_value = feature[1][1]
        analysis_type = feature[1][2]
        if analysis_type == 'lifehistory_based':

            selected_rows = df.drop_duplicates(subset='id')[[feature_related_col_name]].copy()
            selected_rows = selected_rows.loc[~ selected_rows[feature_related_col_name].isna()]

            if feature_related_col_name == 'LifeHistory':

                selected_rows['LifeHistory'] *= interval_time

                # filter
                selected_rows = selected_rows.loc[selected_rows['LifeHistory'] >= doubling_time]

                sel_feature_frequency = calculate_frequency(selected_rows[feature_related_col_name].values,
                                                            feature_related_col_name)
            else:
                sel_feature_frequency = calculate_frequency(selected_rows[feature_related_col_name].values,
                                                            feature_related_col_name)
        elif analysis_type == 'family_based':

            selected_rows = df.drop_duplicates(subset=label_col)[[feature_related_col_name]].copy()
            selected_rows = selected_rows.loc[~ selected_rows[feature_related_col_name].isna()]
            sel_feature_frequency = calculate_frequency(selected_rows[feature_related_col_name].values,
                                                        feature_related_col_name)

        else:

            selected_rows = df.loc[~ df[feature_related_col_name].isna()]
            sel_feature_frequency = calculate_frequency(selected_rows[feature_related_col_name].values,
                                                        feature_related_col_name)

        # Calculate bin edges using numpy.histogram_bin_edges with 'doane'
        bin_edges = np.histogram_bin_edges(sel_feature_frequency[feature_related_col_name], bins='doane')

        # Check if the min or max value falls outside the bin edges
        extend_lower = False
        if sel_feature_frequency[feature_related_col_name].min() < bin_edges[0]:
            # Extend lower
            bin_edges = np.insert(bin_edges, 0, sel_feature_frequency[feature_related_col_name].min() - 1)
            extend_lower = True

        extend_upper = False
        if sel_feature_frequency[feature_related_col_name].max() > bin_edges[-1]:
            # Extend upper
            bin_edges = np.append(bin_edges, sel_feature_frequency[feature_related_col_name].max() + 1)
            extend_upper = True

        # Assign each LifeHistory value to a bin
        sel_feature_frequency["bin"] = pd.cut(sel_feature_frequency[feature_related_col_name], bins=bin_edges,
                                              labels=False, include_lowest=True)

        grouped_sel_feature = sel_feature_frequency.groupby("bin").sum()

        # Ensure all bins are represented
        all_bins = pd.Series(0, index=range(len(bin_edges) - 1))  # Create all possible bins
        grouped_sel_feature = grouped_sel_feature.reindex(all_bins.index, fill_value=0)

        # Create x positions for grouped bars
        x_positions = np.arange(len(grouped_sel_feature.index))

        # Create labels for the bins
        bin_labels = []
        if feature_related_col_name in ['LifeHistory', 'Division_Family_Count']:
            for i in range(len(bin_edges) - 1):
                if i == 0:
                    if extend_lower:
                        bin_labels.append(f"<{bin_edges[1]:.0f}")  # Label for the new lower bin
                    else:
                        bin_labels.append(f"[{bin_edges[i]:.0f},{bin_edges[i + 1]:.0f}]")
                elif i == len(bin_edges) - 2:
                    if extend_upper:
                        bin_labels.append(f">{bin_edges[-2]:.0f}")  # Label for the new upper bin
                    else:
                        bin_labels.append(f"({bin_edges[i]:.0f},{bin_edges[i + 1]:.0f}]")
                else:
                    bin_labels.append(f"({bin_edges[i]:.0f},{bin_edges[i + 1]:.0f}]")
        elif feature_related_col_name in ['Elongation_Rate', 'Velocity']:
            for i in range(len(bin_edges) - 1):
                if i == 0:
                    if extend_lower:
                        bin_labels.append(f"<{bin_edges[1]:.4f}")  # Label for the new lower bin
                    else:
                        bin_labels.append(f"[{bin_edges[i]:.4f},{bin_edges[i + 1]:.4f}]")
                elif i == len(bin_edges) - 2:
                    if extend_upper:
                        bin_labels.append(f">{bin_edges[-2]:.4f}")  # Label for the new upper bin
                    else:
                        bin_labels.append(f"({bin_edges[i]:.4f},{bin_edges[i + 1]:.4f}]")
                else:
                    bin_labels.append(f"({bin_edges[i]:.4f},{bin_edges[i + 1]:.4f}]")
        else:

            for i in range(len(bin_edges) - 1):
                if i == 0:
                    if extend_lower:
                        bin_labels.append(f"<{bin_edges[1]:.2f}")  # Label for the new lower bin
                    else:
                        bin_labels.append(f"[{bin_edges[i]:.2f},{bin_edges[i + 1]:.2f}]")
                elif i == len(bin_edges) - 2:
                    if extend_upper:
                        bin_labels.append(f">{bin_edges[-2]:.2f}")  # Label for the new upper bin
                    else:
                        bin_labels.append(f"({bin_edges[i]:.2f},{bin_edges[i + 1]:.2f}]")
                else:
                    bin_labels.append(f"({bin_edges[i]:.2f},{bin_edges[i + 1]:.2f}]")

        plot_frequency_distribution(x_positions, grouped_sel_feature, bin_labels, feature_related_col_name,
                                    feature_name, y_axis_value, output_path)



import numpy as np


def mark_cell_types_by_intensity(df, cell_type_array, cols, intensity_threshold):

    """
    Marks cell types as present (1) or absent (0) based on intensity threshold values.

    :param pandas.DataFrame df:
        Input DataFrame containing intensity data for each cell type across multiple columns.
    :param numpy.ndarray cell_type_array:
        A 2D numpy array where each row represents a bacterial cell, and each column represents
        a cell type. This array will be updated with binary values (0 or 1) based on the intensity threshold.
    :param list cols:
        List of column names in `df` corresponding to cell types whose intensities need to be checked.
    :param float intensity_threshold:
        Threshold value above which the cell type is considered present (marked as 1).

    :returns:
        numpy.ndarray:

        Updated 2D numpy array with binary marks indicating cell type presence for each bacterial cell.
    """

    for col_idx, col in enumerate(cols):
        condition = df[col] > intensity_threshold
        cell_type_array[condition, col_idx] = 1

    return cell_type_array


def get_fluorescence_intensity_columns(dataframe_col):

    """
    Retrieves a sorted list of fluorescence intensity column names from the provided DataFrame column names.

    :param pandas.Series dataframe_col:
        A Pandas Series containing column names from a DataFrame.

    :returns:
        list:

        A sorted list of column names that contain the substring `Intensity_MeanIntensity_`.
    """

    # fluorescence intensity columns
    fluorescence_intensities_col = \
        sorted(dataframe_col[dataframe_col.str.contains('Intensity_MeanIntensity_')].values.tolist())

    return fluorescence_intensities_col


def assign_cell_types(dataframe, intensity_threshold):
    """
    Assigns binary cell type labels based on fluorescence intensity thresholds.

    :param pandas.DataFrame dataframe:
        Input DataFrame containing intensity data and other features.
    :param float intensity_threshold:
        Threshold value above which a cell type is marked as present (1).

    :returns:
        numpy.ndarray:

        A 2D numpy array where each row represents a cell, and each column represents a cell type, marked as 1 or 0
        based on intensity threshold.
    """

    # If the CSV file has two mean intensity columns, the cosine similarity is calculated
    intensity_col_names = get_fluorescence_intensity_columns(dataframe.columns)

    # cell type
    if len(intensity_col_names) >= 1:

        cell_type_array = np.zeros((dataframe.shape[0], len(intensity_col_names)))
        # check  fluorescence intensity
        cell_type_array = mark_cell_types_by_intensity(dataframe, cell_type_array, intensity_col_names,
                                                       intensity_threshold)
    else:
        cell_type_array = np.zeros((dataframe.shape[0], 1))

    return cell_type_array


def determine_final_cell_type(dataframe, cell_type_array):

    """
    Determines and assigns the final cell type classification for each bacterium based on fluorescence intensity data.

    :param pandas.DataFrame dataframe:
        Input DataFrame containing bacterial data and fluorescence intensity columns.
    :param numpy.ndarray cell_type_array:
        A 2D numpy array where each row represents a bacterial cell, and each column represents
        a binary classification (1 or 0) of a specific cell type based on intensity thresholds.

    :returns:
        pandas.DataFrame:
        Updated DataFrame with a new column 'cellType' representing the final cell type
        classification for each bacterium:

            - 3: Assigned if the fluorescence intensities of at least two channels exceed the defined threshold.
            - Specific values (e.g., 2, 3, etc.): Assigned if only one channel’s intensity exceeds the threshold.
            - 0: Assigned if none of the fluorescence intensities across all channels exceed the defined threshold.
            - 1: Default cell type assigned when only one intensity column exists.
    """

    num_intensity_cols = len(get_fluorescence_intensity_columns(dataframe.columns))

    if num_intensity_cols > 1:

        dataframe['cellType'] = 0

        num_1_value_per_bac = np.sum(cell_type_array, axis=1)
        num_0_value_per_bac = cell_type_array.shape[1] - num_1_value_per_bac

        cond1_more_than_2_1_value = num_1_value_per_bac >= 2
        cond2_more_than_2_0_value = num_0_value_per_bac >= 2

        cond3 = ~ cond1_more_than_2_1_value & ~ cond2_more_than_2_0_value

        indices_of_ones_in_columns = np.where(cell_type_array[cond3] == 1)[0]
        cond3_value = indices_of_ones_in_columns + 1

        dataframe.loc[cond1_more_than_2_1_value, 'cellType'] = 3
        dataframe.loc[cond3, 'cellType'] = cond3_value

    else:
        dataframe['cellType'] = 1

    return dataframe



import subprocess
import os


def test_cli_help():
    """Test the --help option of the CLI."""
    result = subprocess.run(
        ["python", "-m", "Trackrefiner.cli", "--help"],
        capture_output=True,
        text=True,
    )
    assert result.returncode == 0  # Ensure the process exits successfully
    assert "analyzing CellProfiler output" in result.stdout  # Check help description


def test_cli_invalid_args():
    """Test the CLI with missing required arguments."""
    result = subprocess.run(
        ["python", "-m", "Trackrefiner.cli"],
        capture_output=True,
        text=True,
    )
    assert result.returncode != 0  # Should fail due to missing required arguments
    assert "usage: " in result.stderr  # Ensure the error mentions usage information


def test_cli_incorrect_data_types():
    """Test the CLI with incorrect data types for numeric arguments."""
    result = subprocess.run(
        [
            "python",
            "-m",
            "Trackrefiner.cli",
            "-i", "test_data/FilterObjects.csv",
            "-s", "test_data/objects/",
            "-n", "test_data/Object_relationships.csv",
            "-t", "not_a_number",  # Invalid interval_time
            "-d", "another_invalid",  # Invalid doubling_time
            "-o", "output",
        ],
        capture_output=True,
        text=True,
    )
    assert result.returncode != 0  # Should fail due to invalid argument types
    assert "Invalid value for interval_time:" in result.stderr or \
           "Invalid value for doubling_time:" in result.stderr, (
        "Expected error message for invalid data types was not found in stderr."
    )


def test_cli_missing_files():
    """Test the CLI with non-existent file paths."""
    result = subprocess.run(
        [
            "python",
            "-m",
            "Trackrefiner.cli",
            "-i", "non_existent_file.csv",  # Non-existent CP output file
            "-s", "non_existent_directory/",  # Non-existent segmentation folder
            "-n", "non_existent_neighbors.csv",  # Non-existent neighbor CSV file
            "-t", "3",
            "-d", "20",
            "-o", "output",
        ],
        capture_output=True,
        text=True,
    )
    assert result.returncode != 0  # Should fail due to missing files
    assert "The CP output CSV file does not exist" in result.stderr or \
           "The neighbor CSV file does not exist" in result.stderr, (
        "Expected error message for missing files was not found in stderr."
    )


def test_cli_disable_tracking_correction(tmp_path):
    """Test the CLI with valid arguments."""
    # Create a temporary output directory
    output_dir = tmp_path / "output"
    os.makedirs(output_dir, exist_ok=True)

    # Simulate running the CLI with valid arguments
    result = subprocess.run(
        [
            "python",
            "-m",
            "Trackrefiner.cli",
            "-i", "test_data/FilterObjects.csv",
            "-n", "test_data/Object_relationships.csv",
            "-t", "3",
            "-d", "20",
            "--disable_tracking_correction",  # Tracking correction disabled
            "-o", str(output_dir),
        ],
        capture_output=True,
        text=True,
    )

    assert result.returncode == 0  # Ensure the process exits successfully
    # Check for the final success log message in stdout
    assert "Trackrefiner Process completed at:" in result.stdout, (
        "Expected log message indicating successful completion was not found in stdout."
    )
    assert output_dir.exists()  # Check if the output directory was created


def test_cli_empty_cp_output(tmp_path):
    """Test the CLI with empty input files."""
    # Create empty test files
    # Create a temporary CP output CSV with only headers
    invalid_cp_output_csv = tmp_path / "invalid_cp_output.csv"
    # Only headers, no data
    invalid_cp_output_csv.write_text(
        "ImageNumber,ObjectNumber,AreaShape_MajorAxisLength,AreaShape_MinorAxisLength,AreaShape_Orientation\n")

    empty_neighbor_csv = tmp_path / "empty_neighbors.csv"
    empty_neighbor_csv.write_text("")  # Empty file

    # Create a temporary output directory
    output_dir = tmp_path / "output"
    os.makedirs(output_dir, exist_ok=True)

    # Simulate running the CLI with empty input files
    result = subprocess.run(
        [
            "python",
            "-m",
            "Trackrefiner.cli",
            "-i", str(invalid_cp_output_csv),
            "-n", "test_data/Object_relationships.csv",
            "-t", "3",
            "-d", "20",
            "-o", str(output_dir),
        ],
        capture_output=True,
        text=True,
    )
    assert result.returncode != 0  # Should fail due to empty input files
    assert f"The {invalid_cp_output_csv} is empty." in result.stderr, (
        "Expected error message for empty input files was not found in stderr."
    )


def test_cli_with_args(tmp_path):
    """Test the CLI with valid arguments."""
    # Create a temporary output directory
    output_dir = tmp_path / "output"
    os.makedirs(output_dir, exist_ok=True)

    # Simulate running the CLI with valid arguments
    result = subprocess.run(
        [
            "python",
            "-m",
            "Trackrefiner.cli",
            "-i", "test_data/FilterObjects.csv",
            "-s", "test_data/objects/",
            "-n", "test_data/Object_relationships.csv",
            "-t", "3",
            "-d", "20",
            "-o", str(output_dir),
        ],
        capture_output=True,
        text=True,
    )

    assert result.returncode == 0  # Ensure the process exits successfully
    # Check for the final success log message in stdout
    assert "Trackrefiner Process completed at:" in result.stdout, (
        "Expected log message indicating successful completion was not found in stdout."
    )
    assert output_dir.exists()  # Check if the output directory was created



from Trackrefiner.core.correction.action.helper import calc_normalized_angle_between_motion_for_df


def calc_motion_alignment_angle_ml(df, neighbor_df, center_coord_cols, selected_rows=None, col_target=None,
                                   col_source=None):

    """
    Calculates the motion alignment angle between the movement direction of a link (source-to-target)
    and the average motion direction of neighboring bacteria. The alignment is computed in degrees,
    representing the angular difference between the link's motion and its neighbors' average motion.

    :param pandas.DataFrame df:
        The main DataFrame containing bacterial data.
    :param pandas.DataFrame neighbor_df:
        A DataFrame containing neighbor relationships (e.g., First and Second Image/Object Numbers).
    :param dict center_coord_cols:
        A dictionary specifying the column names for the x and y coordinates of bacterial centroids
        (e.g., `{"x": "Center_X", "y": "Center_Y"}`).
    :param pandas.DataFrame selected_rows:
        The subset of rows for which motion alignment angles need to be computed.
    :param str col_target:
        Suffix for target-related columns in `selected_rows`.
    :param str col_source:
        Suffix for source-related columns in `selected_rows`.

    **Returns**:
        pandas.DataFrame:

        The updated `selected_rows` DataFrame with a new column:
            - `"Motion_Alignment_Angle" + col_target`: The calculated motion alignment angle (degrees).
    """

    temp_df = df[['ImageNumber', 'ObjectNumber']].copy()

    temp_df['source_target_TrajectoryX'] = df.groupby('id')["TrajectoryX"].shift(-1)
    temp_df['source_target_TrajectoryY'] = df.groupby('id')["TrajectoryY"].shift(-1)

    daughters_idx = ~ df['Total_Daughter_Mother_Length_Ratio'].isna()

    temp_df.loc[daughters_idx, ['source_target_TrajectoryX', 'source_target_TrajectoryY']] = \
        df.loc[daughters_idx, ['Daughter_Avg_TrajectoryX', 'Daughter_Avg_TrajectoryY']].values

    selected_rows['this_link_dir_x'] = (selected_rows[center_coord_cols['x'] + col_target] -
                                        selected_rows[center_coord_cols['x'] + col_source])

    selected_rows['this_link_dir_y'] = (selected_rows[center_coord_cols['y'] + col_target] -
                                        selected_rows[center_coord_cols['y'] + col_source])

    source_bac_neighbors = \
        selected_rows[['ImageNumber' + col_source, 'ObjectNumber' + col_source,
                       'ImageNumber' + col_target, 'ObjectNumber' + col_target,
                       'this_link_dir_x', 'this_link_dir_y']].merge(
            neighbor_df, left_on=['ImageNumber' + col_source, 'ObjectNumber' + col_source],
            right_on=['First Image Number', 'First Object Number'],
            how='left')[['ImageNumber' + col_source, 'ObjectNumber' + col_source,
                         'ImageNumber' + col_target, 'ObjectNumber' + col_target,
                         'this_link_dir_x', 'this_link_dir_y', 'Second Image Number', 'Second Object Number']]

    # be careful: 'source_target_TrajectoryX', 'source_target_TrajectoryY' is only for neighbors
    source_bac_neighbors_info = \
        source_bac_neighbors.merge(
            temp_df, left_on=['Second Image Number', 'Second Object Number'],
            right_on=['ImageNumber', 'ObjectNumber'], how='left',
            suffixes=('', '_neighbor2'))[['ImageNumber' + col_source, 'ObjectNumber' + col_source,
                                          'ImageNumber' + col_target, 'ObjectNumber' + col_target,
                                          'this_link_dir_x', 'this_link_dir_y',
                                          'source_target_TrajectoryX', 'source_target_TrajectoryY']]

    source_bac_neighbors_info['neighbors_avg_dir_motion_x'] = \
        source_bac_neighbors_info.groupby(['ImageNumber' + col_target,
                                           'ObjectNumber' + col_target,
                                           'ImageNumber' + col_source,
                                           'ObjectNumber' + col_source])['source_target_TrajectoryX'].transform('mean')

    source_bac_neighbors_info['neighbors_avg_dir_motion_y'] = \
        source_bac_neighbors_info.groupby(['ImageNumber' + col_target,
                                           'ObjectNumber' + col_target,
                                           'ImageNumber' + col_source,
                                           'ObjectNumber' + col_source])['source_target_TrajectoryY'].transform('mean')

    source_bac_neighbors_info = source_bac_neighbors_info.groupby(['ImageNumber' + col_target,
                                                                   'ObjectNumber' + col_target,
                                                                   'ImageNumber' + col_source,
                                                                   'ObjectNumber' + col_source
                                                                   ]).head(1)

    source_bac_neighbors_info = \
        calc_normalized_angle_between_motion_for_df(
            source_bac_neighbors_info, 'this_link_dir_x', 'this_link_dir_y',
            'neighbors_avg_dir_motion_x', 'neighbors_avg_dir_motion_y')

    selected_rows = \
        selected_rows.merge(source_bac_neighbors_info[
                                ['ImageNumber' + col_target, 'ObjectNumber' + col_target,
                                 'ImageNumber' + col_source, 'ObjectNumber' + col_source, 'angle_degrees']],
                            on=['ImageNumber' + col_target, 'ObjectNumber' + col_target,
                                'ImageNumber' + col_source, 'ObjectNumber' + col_source], how='inner')

    selected_rows["Motion_Alignment_Angle" + col_target] = selected_rows['angle_degrees']
    selected_rows["Motion_Alignment_Angle" + col_target] = selected_rows["Motion_Alignment_Angle" + col_target].fillna(0)
    selected_rows.drop(columns=['angle_degrees'], inplace=True)

    return selected_rows



import numpy as np
from Trackrefiner.core.correction.action.findOutlier import detect_daughter_to_mother_length_outliers, \
    calculate_length_change_ratio_boundary, calculate_lower_statistical_bound
from Trackrefiner.core.correction.modelTraning.calculation.iouCalForML import calculate_iou_ml
from Trackrefiner.core.correction.modelTraning.calculation.calcDistanceForML import calc_min_distance_ml
from Trackrefiner.core.correction.modelTraning.calculation.lengthRatio import check_len_ratio
from Trackrefiner.core.correction.modelTraning.calculation.calMotionAlignmentAngle import calc_motion_alignment_angle_ml
from Trackrefiner.core.correction.action.neighborAnalysis import compare_neighbor_sets
from Trackrefiner.core.correction.action.helper import calc_neighbors_dir_motion_all, calc_movement


def detect_and_resolve_redundant_parent_link(df, neighbor_df, neighbor_list_array, parent_image_number_col,
                                             parent_object_number_col, center_coord_cols, continuity_links_model,
                                             coordinate_array):

    """
    Detects and resolves redundant parent links (RPL) in bacterial lineage tracking by evaluating mother-daughter
    relationships based on geometric, motion, and neighbor-based features.
    This function performs iterative checks to detect RPLs.

    :param pandas.DataFrame df:
        DataFrame containing tracking data for bacteria.
    :param pandas.DataFrame neighbor_df:
        DataFrame containing neighbor relationships between bacteria.
    :param lil_matrix neighbor_list_array:
        Matrix representing neighbor connections between bacteria.
    :param str parent_image_number_col:
        Column name in `df` for the parent bacterium's image number.
    :param str parent_object_number_col:
        Column name in `df` for the parent bacterium's object number.
    :param dict center_coord_cols:
        Dictionary specifying column names for bacterial centroid coordinates
        (e.g., `{"x": "Center_X", "y": "Center_Y"}`).
    :param sklearn.Model continuity_links_model:
        Machine learning model used to evaluate which target bacterium the source bacterium should link to in order to
         continue its life history.
    :param csr_matrix coordinate_array:
        Matrix of spatial coordinates used for geometric and spatial calculations.

    **Returns**:
        pandas.DataFrame:

        Updated DataFrame with resolved RPLs and recalculated features.
    """

    num_redundant_links = None

    while num_redundant_links != 0 and len(df["Total_Daughter_Mother_Length_Ratio"].dropna().values) > 0:

        wrong_daughters_index_list = []
        mother_with_rpl_list = []
        unexpected_end_mothers = []

        incorrect_link_target_list = []

        # note: it's only for division (mother - daughters relation)
        # check daughter length (sum daughters length or max daughter length) to mother length
        bacteria_with_redundant_parent_link_error = detect_daughter_to_mother_length_outliers(df)
        num_redundant_links = bacteria_with_redundant_parent_link_error.shape[0]

        if num_redundant_links != 0:

            bac_len_to_bac_ratio_boundary = calculate_length_change_ratio_boundary(df)

            lower_bound_threshold = calculate_lower_statistical_bound(bac_len_to_bac_ratio_boundary)

            # rpl mother with daughters
            source_bac_with_rpl = \
                bacteria_with_redundant_parent_link_error.merge(df, left_on=['ImageNumber', 'ObjectNumber'],
                                                                right_on=[parent_image_number_col,
                                                                          parent_object_number_col], how='inner',
                                                                suffixes=('_bac1', ''))

            col_target = ''
            col_source = '_bac1'

            # index correction
            source_bac_with_rpl['index_prev' + col_target] = \
                source_bac_with_rpl['index' + col_target]
            source_bac_with_rpl['index2' + col_target] = source_bac_with_rpl.index.values

            # check incorrect links
            incorrect_links = \
                source_bac_with_rpl.loc[source_bac_with_rpl['Length_Change_Ratio' + col_target] < lower_bound_threshold]

            # also this
            incorrect_link_target_list.extend(incorrect_links['index_prev' + col_target].values.tolist())

            # calculate features & apply model

            source_bac_with_rpl = calculate_iou_ml(source_bac_with_rpl, col_source='prev_index' + col_source,
                                                   col_target='prev_index' + col_target, link_type='continuity',
                                                   coordinate_array=coordinate_array)

            source_bac_with_rpl = \
                calc_min_distance_ml(source_bac_with_rpl, center_coord_cols, postfix_target=col_target,
                                     postfix_source=col_source, link_type=None)

            source_bac_with_rpl['Neighbor_Difference_Count' + col_target] = np.nan
            source_bac_with_rpl['other_daughter_index' + col_target] = np.nan
            source_bac_with_rpl['parent_id' + col_target] = source_bac_with_rpl['id' + col_source]
            source_bac_with_rpl['prev_time_step_index' + col_target] = \
                source_bac_with_rpl['prev_time_step_index' + col_target].astype('int32')

            source_bac_with_rpl = \
                compare_neighbor_sets(source_bac_with_rpl, neighbor_list_array,
                                      parent_image_number_col, parent_object_number_col,
                                      selected_rows_df=source_bac_with_rpl, selected_time_step_df=df,
                                      return_common_elements=True, col_target=col_target)

            source_bac_with_rpl = check_len_ratio(df, source_bac_with_rpl, col_target=col_target,
                                                  col_source=col_source)

            # motion alignment
            # calculated for original df and we should calc for new df
            source_bac_with_rpl["Motion_Alignment_Angle" + col_target] = np.nan
            source_bac_with_rpl = \
                calc_motion_alignment_angle_ml(df, neighbor_df, center_coord_cols,
                                               selected_rows=source_bac_with_rpl, col_target=col_target,
                                               col_source=col_source)

            source_bac_with_rpl['adjusted_Neighbor_Shared_Count' + col_target] = np.where(
                source_bac_with_rpl['Neighbor_Shared_Count' + col_target] == 0,
                source_bac_with_rpl['Neighbor_Shared_Count' + col_target] + 1,
                source_bac_with_rpl['Neighbor_Shared_Count' + col_target]
            )

            source_bac_with_rpl['neighbor_ratio' + col_target] = \
                (source_bac_with_rpl['Neighbor_Difference_Count' + col_target] / (
                    source_bac_with_rpl['adjusted_Neighbor_Shared_Count' + col_target]))

            raw_feature_list = ['iou', 'min_distance', 'Neighbor_Difference_Count' + col_target,
                                'Neighbor_Shared_Count' + col_target,
                                'length_dynamic' + col_target,
                                'Motion_Alignment_Angle' + col_target,
                                'neighbor_ratio' + col_target,
                                'index' + col_source, 'index_prev' + col_target]

            source_bac_with_rpl = source_bac_with_rpl[raw_feature_list].copy()
            source_bac_with_rpl = source_bac_with_rpl.rename(
                {
                    'Neighbor_Shared_Count' + col_target: 'Neighbor_Shared_Count',
                    'neighbor_ratio' + col_target: 'neighbor_ratio',
                    'Neighbor_Difference_Count' + col_target: 'Neighbor_Difference_Count',
                    'length_dynamic' + col_target: 'length_dynamic',
                    'Motion_Alignment_Angle' + col_target: 'Motion_Alignment_Angle',
                    'index_prev' + col_target: 'index_prev',
                }, axis=1)

            # difference_neighbors
            feature_list_for_non_divided_bac_model = \
                ['iou', 'min_distance', 'neighbor_ratio', 'length_dynamic', 'Motion_Alignment_Angle']

            y_prob_non_divided_bac_model = \
                continuity_links_model.predict_proba(
                    source_bac_with_rpl[feature_list_for_non_divided_bac_model])[:, 1]

            source_bac_with_rpl['prob_non_divided_bac_model'] = y_prob_non_divided_bac_model

            # Pivot this DataFrame to get the desired structure
            continuity_link_cost_df = \
                source_bac_with_rpl[['index' + col_source, 'index_prev', 'prob_non_divided_bac_model']].pivot(
                    index='index' + col_source, columns='index_prev', values='prob_non_divided_bac_model')

            continuity_link_cost_df.columns.name = None
            continuity_link_cost_df.index.name = None

            continuity_link_cost_df = 1 - continuity_link_cost_df

            for parent_with_redundant_link_idx, this_rpl_cost in continuity_link_cost_df.iterrows():

                daughters_idx_of_this_mother = this_rpl_cost[this_rpl_cost.notna()].index.tolist()
                parent_with_redundant_link = df.loc[parent_with_redundant_link_idx]

                incorrect_daughters_detected_in_prev_conditions = [v for v in daughters_idx_of_this_mother if v in
                                                                   incorrect_link_target_list]

                if len(incorrect_daughters_detected_in_prev_conditions) == 1:

                    # it means that one daughter is incorrect and the other is correct
                    correct_daughter = [v for v in daughters_idx_of_this_mother if v not in
                                        incorrect_daughters_detected_in_prev_conditions][0]

                    df.loc[df['id'] == df.at[incorrect_daughters_detected_in_prev_conditions[0], 'id'], 'parent_id'] = 0
                    wrong_daughters_index_list.append(incorrect_daughters_detected_in_prev_conditions[0])

                    correct_daughter_info = df.loc[correct_daughter]

                    # change feature of prev parent
                    df.loc[df['id'] == parent_with_redundant_link['id'], 'divideFlag'] = \
                        df.at[correct_daughter, 'divideFlag']

                    df.loc[df['id'] == parent_with_redundant_link['id'], 'Division_TimeStep'] = \
                        df.at[correct_daughter, 'Division_TimeStep']

                    df.at[correct_daughter, 'Length_Change_Ratio'] = \
                        df.at[correct_daughter, 'Daughter_Mother_Length_Ratio']

                    bac_movement = calc_movement(correct_daughter_info, parent_with_redundant_link, center_coord_cols)

                    df.at[correct_daughter, 'Bacterium_Movement'] = bac_movement

                    df.at[correct_daughter, 'other_daughter_index'] = np.nan
                    df.at[correct_daughter, 'parent_index'] = np.nan
                    df.at[correct_daughter, 'Daughter_Mother_Length_Ratio'] = np.nan

                    df.at[correct_daughter, 'Prev_MajorAxisLength'] = \
                        parent_with_redundant_link['AreaShape_MajorAxisLength']

                    df.at[correct_daughter, 'Prev_Center_X'] = \
                        parent_with_redundant_link[center_coord_cols['x']]
                    df.at[correct_daughter, 'Prev_Center_Y'] = \
                        parent_with_redundant_link[center_coord_cols['y']]

                    df.at[correct_daughter, 'Prev_Endpoint1_X'] = parent_with_redundant_link['Endpoint1_X']
                    df.at[correct_daughter, 'Prev_Endpoint1_Y'] = parent_with_redundant_link['Endpoint1_Y']
                    df.at[correct_daughter, 'Prev_Endpoint2_X'] = parent_with_redundant_link['Endpoint2_X']
                    df.at[correct_daughter, 'Prev_Endpoint2_Y'] = parent_with_redundant_link['Endpoint2_Y']

                    df.loc[df['parent_id'] == correct_daughter_info['id'], 'parent_id'] = \
                        parent_with_redundant_link['id']
                    df.loc[df['id'] == correct_daughter_info['id'], 'parent_id'] = \
                        parent_with_redundant_link['parent_id']
                    df.loc[df['id'] == correct_daughter_info['id'], 'id'] = parent_with_redundant_link['id']

                    # now we should change features of prev parent bacterium
                    df.loc[parent_with_redundant_link_idx, ['daughters_index', 'Total_Daughter_Mother_Length_Ratio',
                                                            'Max_Daughter_Mother_Length_Ratio', 'Daughter_Avg_TrajectoryX',
                                                            'Daughter_Avg_TrajectoryY']] = \
                        [np.nan, np.nan, np.nan, np.nan, np.nan, ]

                elif len(incorrect_daughters_detected_in_prev_conditions) == 0:

                    # it means both daughters pass the prev conditions
                    # cost is equal to 1 - probability
                    incorrect_daughter = this_rpl_cost.idxmax()
                    # incorrect_daughter_prob = this_rpl_cost.max()

                    correct_daughter = this_rpl_cost.idxmin()
                    correct_daughter_prob = this_rpl_cost.min()

                    if incorrect_daughter == correct_daughter:
                        incorrect_daughter = [v for v in this_rpl_cost[this_rpl_cost.notna()].index.values if
                                              v != correct_daughter][0]

                    if correct_daughter_prob <= 0.5:

                        # it means one daughter is incorrect and one daughter is correct
                        df.loc[df['id'] == df.at[incorrect_daughter, 'id'], 'parent_id'] = 0
                        wrong_daughters_index_list.append(incorrect_daughter)

                        # now we should calculate features value of correct bac
                        correct_daughter_info = df.loc[correct_daughter]

                        # change feature of prev parent
                        df.loc[df['id'] == parent_with_redundant_link['id'], 'divideFlag'] = \
                            df.at[correct_daughter, 'divideFlag']

                        df.loc[df['id'] == parent_with_redundant_link['id'], 'Division_TimeStep'] = \
                            df.at[correct_daughter, 'Division_TimeStep']

                        df.at[correct_daughter, 'Length_Change_Ratio'] = \
                            df.at[correct_daughter, 'Daughter_Mother_Length_Ratio']

                        bac_movement = \
                            calc_movement(correct_daughter_info, parent_with_redundant_link, center_coord_cols)
                        df.at[correct_daughter, 'Bacterium_Movement'] = bac_movement

                        df.at[correct_daughter, 'other_daughter_index'] = np.nan
                        df.at[correct_daughter, 'parent_index'] = np.nan
                        df.at[correct_daughter, 'Daughter_Mother_Length_Ratio'] = np.nan

                        df.at[correct_daughter, 'Prev_MajorAxisLength'] = \
                            parent_with_redundant_link['AreaShape_MajorAxisLength']

                        df.at[correct_daughter, 'Prev_Center_X'] = \
                            parent_with_redundant_link[center_coord_cols['x']]
                        df.at[correct_daughter, 'Prev_Center_Y'] = \
                            parent_with_redundant_link[center_coord_cols['y']]

                        df.at[correct_daughter, 'Prev_Endpoint1_X'] = parent_with_redundant_link[
                            'Endpoint1_X']
                        df.at[correct_daughter, 'Prev_Endpoint1_Y'] = parent_with_redundant_link[
                            'Endpoint1_Y']
                        df.at[correct_daughter, 'Prev_Endpoint2_X'] = parent_with_redundant_link[
                            'Endpoint2_X']
                        df.at[correct_daughter, 'Prev_Endpoint2_Y'] = parent_with_redundant_link[
                            'Endpoint2_Y']

                        df.loc[df['parent_id'] == correct_daughter_info['id'], 'parent_id'] = \
                            parent_with_redundant_link['id']
                        df.loc[df['id'] == correct_daughter_info['id'], 'parent_id'] = parent_with_redundant_link[
                            'parent_id']
                        df.loc[df['id'] == correct_daughter_info['id'], 'id'] = parent_with_redundant_link['id']

                        # now we should change features of prev parent bacterium
                        df.loc[parent_with_redundant_link_idx, ['daughters_index', 'Total_Daughter_Mother_Length_Ratio',
                                                                'Max_Daughter_Mother_Length_Ratio',
                                                                'Daughter_Avg_TrajectoryX',
                                                                'Daughter_Avg_TrajectoryY']] = \
                            [np.nan, np.nan, np.nan, np.nan, np.nan, ]

                    else:
                        # it means the probability of both daughters are low
                        df.loc[df['id'] == df.at[incorrect_daughter, 'id'], 'parent_id'] = 0
                        df.loc[df['id'] == df.at[correct_daughter, 'id'], 'parent_id'] = 0

                        unexpected_end_mothers.append(parent_with_redundant_link_idx)
                        wrong_daughters_index_list.extend([incorrect_daughter, correct_daughter])

                elif len(incorrect_daughters_detected_in_prev_conditions) == 2:

                    daughter_1_idx = incorrect_daughters_detected_in_prev_conditions[0]
                    daughter_2_idx = incorrect_daughters_detected_in_prev_conditions[1]

                    # it's challenging and means both daughters are incorrect based on conditions
                    df.loc[df['id'] == df.at[daughter_1_idx, 'id'], 'parent_id'] = 0
                    df.loc[df['id'] == df.at[daughter_2_idx, 'id'], 'parent_id'] = 0

                    unexpected_end_mothers.append(parent_with_redundant_link_idx)
                    wrong_daughters_index_list.extend([daughter_1_idx, daughter_2_idx])

            if len(wrong_daughters_index_list) > 0:
                df.loc[wrong_daughters_index_list, [parent_image_number_col, parent_object_number_col,
                                                    'Unexpected_Beginning', 'Direction_of_Motion', "TrajectoryX",
                                                    "TrajectoryY", 'prev_time_step_index',
                                                    'other_daughter_index', 'Prev_Bacterium_Slope', 'Orientation_Angle_Between_Slopes',
                                                    'parent_index', 'Daughter_Mother_Length_Ratio']] = \
                    [0, 0, True, np.nan, np.nan, np.nan, -1, np.nan, np.nan, np.nan, np.nan, np.nan]

                df.loc[mother_with_rpl_list, ["Total_Daughter_Mother_Length_Ratio", "Max_Daughter_Mother_Length_Ratio",
                                              'Daughter_Avg_TrajectoryX', 'Daughter_Avg_TrajectoryY']] = \
                    [np.nan, np.nan, np.nan, np.nan]

            if len(unexpected_end_mothers) > 0:
                df.loc[unexpected_end_mothers, ['daughters_index', 'Total_Daughter_Mother_Length_Ratio',
                                                'Max_Daughter_Mother_Length_Ratio', 'Daughter_Avg_TrajectoryX',
                                                'Daughter_Avg_TrajectoryY', 'Unexpected_End']] = \
                    [np.nan, np.nan, np.nan, np.nan, np.nan, True]

    # note: by removing rpl, neighbor set doesn't change (different or common)
    # set age
    df['age'] = df.groupby('id').cumcount() + 1
    df['LifeHistory'] = df.groupby('id')['id'].transform('size')
    # average length of a bacteria in its life history
    df['Avg_Length_Change_Ratio'] = df.groupby('id')['Length_Change_Ratio'].transform('mean')

    # check division
    # _2: bac2, _1: bac1 (source bac)
    merged_df = df.merge(df, left_on=[parent_image_number_col, parent_object_number_col],
                         right_on=['ImageNumber', 'ObjectNumber'], how='inner', suffixes=('_2', '_1'))

    division_df = merged_df[merged_df.duplicated(subset='index_1', keep=False)][['ImageNumber_1', 'ObjectNumber_1',
                                                                                 'index_1', 'id_1', 'index_2',
                                                                                 'ImageNumber_2', 'ObjectNumber_2',
                                                                                 'AreaShape_MajorAxisLength_1',
                                                                                 'AreaShape_MajorAxisLength_2',
                                                                                 'Bacterium_Slope_1', 'Bacterium_Slope_2',
                                                                                 center_coord_cols['x'] + '_1',
                                                                                 center_coord_cols['y'] + '_1',
                                                                                 center_coord_cols['x'] + '_2',
                                                                                 center_coord_cols['y'] + '_2',
                                                                                 parent_object_number_col + '_2',
                                                                                 parent_image_number_col + '_2']].copy()
    # motion alignment
    df = calc_neighbors_dir_motion_all(df, neighbor_df, division_df)

    # check neighbors
    df = compare_neighbor_sets(df, neighbor_list_array, parent_image_number_col, parent_object_number_col)

    df['ovd_flag'] = False
    df['bad_daughters_flag'] = False
    df['bad_division_flag'] = False
    df['daughter_rpl'] = False
    df['mother_rpl'] = False

    return df



import pandas as pd
from Trackrefiner.core.correction.action.bacteriaTrackingUpdate import bacteria_modification
from Trackrefiner.core.correction.action.trackingCost.calculateCreateLinkCost import optimize_assignment_using_hungarian
from Trackrefiner.core.correction.action.trackingCost.calculateCreateLinkCost import calc_division_link_cost_for_restoring_links, \
    calc_continuity_link_cost_for_restoring_links


def add_tracking_link(df, neighbors_df, neighbor_list_array, source_bac_idx, target_bac_idx,
                      parent_image_number_col, parent_object_number_col, center_coord_cols,
                      all_bac_in_target_bac_time_step, prob_val):

    """
    Adds a new tracking link between a source bacterium and a target bacterium if conditions are met.

    :param pandas.DataFrame df:
        The main DataFrame containing tracking information for bacteria.
    :param pandas.DataFrame neighbors_df:
        DataFrame representing the neighbor relationships between bacteria.
    :param scipy.sparse.csr_matrix neighbor_list_array:
        Sparse matrix representing neighborhood connectivity for all bacteria.
    :param int source_bac_idx:
        Index of the source bacterium in the current DataFrame.
    :param int target_bac_idx:
        Index of the target bacterium in the current DataFrame.
    :param str parent_image_number_col:
        Column name for the parent bacterium's image number.
    :param str parent_object_number_col:
        Column name for the parent bacterium's object number.
    :param dict center_coord_cols:
        Dictionary specifying the column names for the x and y centroid coordinates,
        e.g., `{'x': 'Center_X', 'y': 'Center_Y'}`.
    :param pandas.DataFrame all_bac_in_target_bac_time_step:
        Subset of the DataFrame containing all bacteria in the target bacterium's time step.
    :param float prob_val:
        Probability value indicating the likelihood of the new link being valid. A lower probability (`1 - prob_val`)
        must exceed a threshold (0.5) for the link to be considered.

    **Returns**:
        pandas.DataFrame:
            The updated DataFrame with the new tracking link added if the conditions are met.
    """

    if 1 - prob_val > 0.5:
        # there is two scenario: first: source bac with only one bac in next time step: so we compare
        # the probability of that with new bac
        # the second scenario is source bac has two daughters. I think it's good idea two compare
        # max daughter probability with new bac probability

        # update info
        source_bac = df.loc[source_bac_idx]
        # unexpected beginning
        target_bac_life_history = df.loc[df['id'] == df.at[target_bac_idx, 'id']]

        df = bacteria_modification(df, source_bac, target_bac_life_history,
                                   all_bac_in_target_bac_time_step,
                                   neighbors_df, neighbor_list_array, parent_image_number_col, parent_object_number_col,
                                   center_coord_cols)

    return df


def restore_tracking_links(df, neighbors_df, neighbor_list_array, parent_image_number_col, parent_object_number_col,
                           center_coord_cols, df_raw_before_rpl_errors, continuity_links_model, division_links_model,
                           coordinate_array):

    """
    Attempts to restore previously removed tracking links for bacteria by re-evaluating conditions and costs.
    This function focuses on resolving links for unexpected beginning bacteria while maintaining tracking integrity.

    **Process**:
        - Identifies bacteria with `unexpected_beginning` status.
        - Cross-checks these bacteria against the raw tracking data to find potential links that were removed.
        - Evaluates potential links based on:
            - Continuity (single-link bacteria).
            - Division (bacteria dividing into daughters).
        - Applies machine learning models to calculate costs for restoring links and uses
          an optimization algorithm (Hungarian method) to finalize the link restoration.
        - Updates the DataFrame with restored links if conditions are satisfied.

    **Key Features**:
        - **Continuity Links**:
            - Checks whether the bacterium in question can continue its life history based on its source link.
        - **Division Links**:
            - Ensures division links meet criteria, such as the daughter-to-mother length ratio being below a threshold.
        - **Optimization**:
            - Uses a cost matrix and optimization algorithm to assign links with minimal cost.

    :param pandas.DataFrame df:
        The current DataFrame containing bacteria tracking data.
    :param pandas.DataFrame neighbors_df:
        DataFrame representing the neighbor relationships between bacteria.
    :param lil_matrix neighbor_list_array:
        Sparse matrix representing neighborhood connectivity for all bacteria.
    :param str parent_image_number_col:
        Column name for the parent bacterium's image number.
    :param str parent_object_number_col:
        Column name for the parent bacterium's object number.
    :param dict center_coord_cols:
        Dictionary containing the column names for x and y centroid coordinates,
        e.g., `{'x': 'Center_X', 'y': 'Center_Y'}`.
    :param pandas.DataFrame df_raw_before_rpl_errors:
        Original tracking data before resolving RPL errors, used to verify previously existing links.
    :param sklearn.Model continuity_links_model:
        Machine learning model to evaluate the validity of non-divided bacteria continuity links.
    :param sklearn.Model division_links_model:
        Machine learning model to evaluate division-based links for bacteria.
    :param numpy.ndarray coordinate_array:
        Array containing precomputed spatial coordinates for bacteria.

    **Returns**:
        pandas.DataFrame:
            The updated DataFrame with restored tracking links where applicable.
    """

    # only we should check unexpected beginning bacteria and compare with previous links to make sure
    # it can be possible to restore links or not

    unexpected_beginning_bacteria = df.loc[df['Unexpected_Beginning'] == True]

    division_df = pd.DataFrame()
    continuity_links_df = pd.DataFrame()

    for unexpected_bac_idx in unexpected_beginning_bacteria.index.values:

        unexpected_bac = df.loc[[unexpected_bac_idx]]

        bac_related_to_this_bac_in_raw_df = \
            df_raw_before_rpl_errors.loc[df_raw_before_rpl_errors['prev_index'] ==
                                         unexpected_bac['prev_index'].values[0]]

        if bac_related_to_this_bac_in_raw_df[parent_image_number_col].values[0] != 0:

            # it means that this bacterium had a previous link
            source_link = df.loc[
                (df['ImageNumber'] == bac_related_to_this_bac_in_raw_df[parent_image_number_col].values[0]) &
                (df['ObjectNumber'] == bac_related_to_this_bac_in_raw_df[parent_object_number_col].values[0])]

            if source_link['Unexpected_End'].values[0]:
                check_prob = True
                link_type = 'continuity'

            else:
                if str(source_link['Total_Daughter_Mother_Length_Ratio'].values[0]) != 'nan':
                    # it means that source bacterium has two links right now, and we can not restore links
                    check_prob = False
                else:
                    # discuses more about min life history for this
                    check_prob = True
                    other_daughter = df.loc[(df['id'] == source_link['id'].values[0]) & (
                            df['ImageNumber'] == unexpected_bac['ImageNumber'].values[0])]
                    link_type = 'div'

            if check_prob:

                df_bac_with_source = unexpected_bac.merge(source_link, how='cross', suffixes=('', '_source'))

                if link_type == 'continuity':
                    continuity_links_df = pd.concat([continuity_links_df, df_bac_with_source], ignore_index=True)

                elif link_type == 'div':
                    max_daughter_len = (max(other_daughter['AreaShape_MajorAxisLength'].values[0],
                                            unexpected_bac['AreaShape_MajorAxisLength'].values[0]) /
                                        source_link['AreaShape_MajorAxisLength'].values[0])
                    if max_daughter_len < 1:
                        division_df = pd.concat([division_df, df_bac_with_source], ignore_index=True)

    if continuity_links_df.shape[0] > 0:
        bac_with_same_source = continuity_links_df[
            continuity_links_df.duplicated(['ImageNumber_source', 'ObjectNumber_source'],
                                           keep=False)]

        if bac_with_same_source.shape[0] > 0:
            # Group by column 2 and aggregate
            agg_df = bac_with_same_source.groupby(['ImageNumber_source', 'ObjectNumber_source']).agg(
                {'AreaShape_MajorAxisLength': 'max', 'AreaShape_MajorAxisLength_source': 'mean'}).reset_index()

            # Rename columns for clarity
            agg_df.columns = ['ImageNumber_source', 'ObjectNumber_source', 'max_daughter_len_same', 'source_len_same']
            agg_df['Max_Daughter_Mother_Length_Ratio_same'] = agg_df['max_daughter_len_same'] / agg_df['source_len_same']
            merged_df = pd.merge(bac_with_same_source, agg_df, on=['ImageNumber_source', 'ObjectNumber_source'],
                                 how='inner')

            # now we should check daughters condition
            daughters_passed_condition = merged_df.loc[merged_df['Max_Daughter_Mother_Length_Ratio_same'] < 1]
            daughters_passed_condition = daughters_passed_condition.drop(columns=['max_daughter_len_same',
                                                                                  'source_len_same'])

            division_df = pd.concat([division_df, daughters_passed_condition], ignore_index=True)

            # now we should remove them from continuity link df
            continuity_links_df = continuity_links_df.loc[~ continuity_links_df.index.isin(
                bac_with_same_source.index.values)]

    if division_df.shape[0] > 0:

        division_cost_df = \
            calc_division_link_cost_for_restoring_links(df, neighbors_df, neighbor_list_array, division_df,
                                                        center_coord_cols, col_source='_source', col_target='',
                                                        parent_image_number_col=parent_image_number_col,
                                                        parent_object_number_col=parent_object_number_col,
                                                        division_links_model=division_links_model,
                                                        coordinate_array=coordinate_array)

        if division_cost_df.shape[0] > 0:

            division_cost_df = division_cost_df.fillna(1)
            # optimization
            optimized_df = optimize_assignment_using_hungarian(division_cost_df)

            for row_index, row in optimized_df.iterrows():
                source_bac_idx = row['idx1']

                target_bac_idx = int(row['idx2'])
                target_bac = df.loc[target_bac_idx]

                cost_val = row['cost']

                all_bac_in_target_bac_time_step = df.loc[df['ImageNumber'] == target_bac['ImageNumber']]

                df = add_tracking_link(df, neighbors_df, neighbor_list_array, source_bac_idx, target_bac_idx,
                                       parent_image_number_col, parent_object_number_col, center_coord_cols,
                                       all_bac_in_target_bac_time_step, cost_val)

    if continuity_links_df.shape[0] > 0:

        continuity_links_df['id_source'] = df.loc[
            df['index'].isin(continuity_links_df['index_source'].values), 'id'].values
        continuity_links_df['age_source'] = df.loc[
            df['index'].isin(continuity_links_df['index_source'].values), 'age'].values

        continuity_cost_df = \
            calc_continuity_link_cost_for_restoring_links(df, neighbors_df, neighbor_list_array, continuity_links_df,
                                                          center_coord_cols, col_source='_source', col_target='',
                                                          parent_image_number_col=parent_image_number_col,
                                                          parent_object_number_col=parent_object_number_col,
                                                          continuity_links_model=continuity_links_model,
                                                          coordinate_array=coordinate_array)

        if continuity_cost_df.shape[0] > 0:

            continuity_cost_df = continuity_cost_df.fillna(1)
            # optimization
            optimized_df = optimize_assignment_using_hungarian(continuity_cost_df)

            for row_index, row in optimized_df.iterrows():
                source_bac_idx = row['idx1']

                target_bac_idx = int(row['idx2'])
                target_bac = df.loc[target_bac_idx]

                cost_val = row['cost']

                all_bac_in_target_bac_time_step = df.loc[df['ImageNumber'] == target_bac['ImageNumber']]

                df = add_tracking_link(df, neighbors_df, neighbor_list_array, source_bac_idx,
                                       target_bac_idx, parent_image_number_col,
                                       parent_object_number_col, center_coord_cols,
                                       all_bac_in_target_bac_time_step, cost_val)
    return df



import numpy as np


def calc_average_velocity(df_life_history, center_coord_cols, interval_time):

    """
    Calculates the velocity of a bacterium based on its life history data.

    :param pandas.DataFrame df_life_history:
        Input DataFrame containing the life history of bacteria
    :param dict center_coord_cols:
        A dictionary with keys 'x' and 'y', specifying the column names for center coordinates.
    :param float interval_time:
        Time interval between consecutive images (in minutes).

    :returns:
        float:

        The calculated velocity. If the bacterium is observed for only one time step, returns NaN.
    """

    life_history_length = df_life_history.shape[0]

    division_x_coordinate = df_life_history[center_coord_cols['x']].values[-1]
    division_y_coordinate = df_life_history[center_coord_cols['y']].values[-1]

    # coordinate of bacteria when they are born
    birth_x_coordinate = df_life_history[center_coord_cols['x']].values[0]
    birth_y_coordinate = df_life_history[center_coord_cols['y']].values[0]

    # this condition checks the life history of bacteria
    # If the bacterium exists only one time step: NaN will be reported.
    if life_history_length > 1:

        # Compute the Euclidean distance of the bacteria from the origin at initial and final positions.
        x1 = np.sqrt(birth_x_coordinate ** 2 + birth_y_coordinate ** 2)
        x2 = np.sqrt(division_x_coordinate ** 2 + division_y_coordinate ** 2)

        # Calculate the velocity over the life history length.
        velocity = round((x2 - x1) / (life_history_length * interval_time), 3)

    else:
        velocity = np.nan

    return velocity



import sys
from PyQt5.QtCore import Qt, QUrl, pyqtSignal, QObject
from PyQt5.QtGui import QDesktopServices
from PyQt5.QtWidgets import QApplication, QWidget, QVBoxLayout, QHBoxLayout, QLabel, QLineEdit, QPushButton, \
    QFileDialog, QComboBox, QSpinBox, QDoubleSpinBox, QCheckBox, QTextEdit
from PyQt5.QtWidgets import QMessageBox
import os
from Trackrefiner import process_objects_data


class OutputStream(QObject):
    output_written = pyqtSignal(str)

    def __init__(self):
        super().__init__()
        self.terminal = sys.__stdout__  # Keep the original stdout

    def write(self, text):
        self.output_written.emit(text)
        self.terminal.write(text)  # Also write to the terminal
        self.terminal.flush()  # Ensure immediate output to the terminal

    def flush(self):
        self.terminal.flush()  # Flush the original stdout


class TrackRefinerGUI(QWidget):
    def __init__(self):
        super().__init__()
        self.init_ui()

        # Redirect stdout to the log output
        self.stdout_stream = OutputStream()
        self.stdout_stream.output_written.connect(self.append_log)
        sys.stdout = self.stdout_stream

    def init_ui(self):
        self.setWindowTitle("Trackrefiner")
        self.layout = QVBoxLayout()

        self.layout.setSpacing(5)  # Consistent spacing between rows
        self.layout.setContentsMargins(10, 10, 10, 10)  # Consistent margins

        # Input file selection for CellProfiler Output CSV
        self.cp_output_csv_file = self.create_csv_file_input()
        self.add_tooltipped_widget("CellProfiler Output CSV File:",
                                   "Path to the CP-generated CSV file containing measured bacterial "
                                   "features and tracking information.", self.cp_output_csv_file)

        # Segmentation Results Folder
        self.segmentation_results_dir = self.create_folder_input()
        self.add_tooltipped_widget("Segmentation Results Folder:",
                                   "Path to folder containing .pickle files generated from "
                                   "segmentation by CellProfiler.", self.segmentation_results_dir)

        # Neighbor CSV
        self.neighbor_csv = self.create_csv_file_input()
        self.add_tooltipped_widget("Neighbor CSV File:", "Path to CSV file containing "
                                                         "bacterial neighbor information.", self.neighbor_csv)

        # Interval time and Doubling time in one row
        self.interval_time = self.create_double_spinbox(0.0, 1440.0, None)
        self.doubling_time_of_bacteria = self.create_double_spinbox(0.0, 1440.0, None)
        self.add_two_part_row(
            "Interval Time (minutes):",
            "Time interval between frames in minutes.",
            self.interval_time,
            "Doubling Time (minutes):",
            "Minimum lifespan of bacteria in minutes.",
            self.doubling_time_of_bacteria
        )

        # Elongation rate method and Pixel per micron in one row
        self.elongation_rate_method = self.create_combobox(["Average", "Linear Regression"])
        self.pixel_per_micron = self.create_double_spinbox(0.0, 10.0, 0.144)
        self.add_two_part_row(
            "Elongation Rate Method:",
            "Method to calculate elongation rate: Average or Linear Regression.",
            self.elongation_rate_method,
            "Pixel Per Micron:",
            "Conversion factor for pixels to micrometers.",
            self.pixel_per_micron
        )

        # Assign cell type and Intensity threshold in one row
        self.assigning_cell_type = self.create_checkbox()
        self.assigning_cell_type.setChecked(False)  # Default to unchecked
        self.assigning_cell_type.stateChanged.connect(self.toggle_intensity_threshold_visibility)

        self.intensity_threshold_layout = QHBoxLayout()
        self.add_tooltipped_widget_inline(
            "Assign Cell Type:",
            "Check to assign cell type to objects.",
            self.assigning_cell_type,
            self.intensity_threshold_layout
        )

        self.intensity_threshold = self.create_double_spinbox(0.0, 1.0, 0.1)
        self.intensity_threshold_label = QLabel("Intensity Threshold:")
        self.intensity_threshold_label.setToolTip("Threshold for cell intensity used in cell type assignment.")

        self.intensity_threshold_layout.addWidget(self.intensity_threshold_label)
        self.intensity_threshold_layout.addWidget(self.intensity_threshold)
        self.intensity_threshold_label.hide()
        self.intensity_threshold.hide()
        self.layout.addLayout(self.intensity_threshold_layout)

        # Classifier and Number of CPUs in one row
        self.clf = self.create_combobox([
            "LogisticRegression", "GaussianProcessClassifier", "C-Support Vector Classifier"
        ])
        self.n_cpu = self.create_spinbox(-1, 64, -1)
        self.add_two_part_row(
            "Classifier:",
            "Classifier for track refining.",
            self.clf,
            "Number of CPUs:",
            "Number of CPUs for parallel processing. Use -1 for all available CPUs.",
            self.n_cpu
        )

        # Boundary limits
        self.boundary_limits = self.create_line_edit()
        self.boundary_limits.setPlaceholderText("e.g., 0, 112, 52, 323 means X: 0–112, Y: 52–323")
        self.add_tooltipped_widget(
            "Boundary Limits:",
            "Define boundary limits to exclude objects outside the image boundary.",
            self.boundary_limits
        )

        # Dynamic boundaries
        self.dynamic_boundaries = self.create_csv_file_input()
        self.add_tooltipped_widget(
            "Dynamic Boundaries:",
            "Define time-dependent boundary limits using a CSV file.",
            self.dynamic_boundaries
        )

        # Disable tracking correction and Verbose Output in one row
        self.disable_tracking_correction = self.create_checkbox()
        self.verbose = self.create_checkbox()
        self.save_pickle = self.create_checkbox()
        self.add_three_part_row(
            "Disable Tracking Correction",
            "Check to disable tracking correction on CellProfiler output.",
            self.disable_tracking_correction,
            "Verbose Output",
            "Check to enable detailed log messages.",
            self.verbose,
            "Save in .pickle format",
            "Check to enable saving results in .pickle format.",
            self.save_pickle
        )

        # Output folder
        self.out_dir = self.create_folder_input()
        self.add_tooltipped_widget("Output Folder:", "Folder to save the output results.",
                                   self.out_dir)

        # Submit button
        self.submit_button = QPushButton("Run")
        self.submit_button.clicked.connect(self.run_analysis)
        self.layout.addWidget(self.submit_button)

        # Help and Issue buttons
        help_issue_layout = QHBoxLayout()

        # Help button
        self.help_button = QPushButton("Help")
        self.help_button.setToolTip("View the Trackrefiner wiki")
        self.help_button.clicked.connect(lambda: QDesktopServices.openUrl(
            QUrl("https://github.com/Ati-74/Trackrefiner")))
        help_issue_layout.addWidget(self.help_button)

        # Issue button
        self.issue_button = QPushButton("Report Issue")
        self.issue_button.setToolTip("Report an issue with the Trackrefiner")
        self.issue_button.clicked.connect(
            lambda: QDesktopServices.openUrl(QUrl("https://github.com/Ati-74/Trackrefiner/issues")))
        help_issue_layout.addWidget(self.issue_button)

        # Add Help and Issue layout
        self.layout.addLayout(help_issue_layout)

        # Log output
        self.log_output = QTextEdit()
        self.log_output.setReadOnly(True)
        self.layout.addWidget(self.log_output)

        self.setLayout(self.layout)

    def toggle_intensity_threshold_visibility(self, state):
        if state == Qt.Checked:
            self.intensity_threshold_label.show()
            self.intensity_threshold.show()
        else:
            self.intensity_threshold_label.hide()
            self.intensity_threshold.hide()

        # Force layout recalculation
        self.layout.invalidate()

    def append_log(self, text):
        self.log_output.moveCursor(self.log_output.textCursor().End)
        self.log_output.insertPlainText(text)

    def closeEvent(self, event):
        sys.stdout = sys.__stdout__  # Restore original stdout
        super().closeEvent(event)

    def add_tooltipped_widget(self, label_text, tooltip_text, widget):
        layout = QHBoxLayout()
        label = QLabel(label_text)
        label.setToolTip(tooltip_text)

        # Adjust layout spacing and alignment
        layout.setSpacing(5)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setAlignment(Qt.AlignLeft)

        layout.addWidget(label)
        layout.addWidget(widget)
        self.layout.addLayout(layout)

    def add_tooltipped_widget_inline(self, label_text, tooltip_text, widget, layout):
        label = QLabel(label_text)
        label.setToolTip(tooltip_text)
        layout.addWidget(label)
        layout.addWidget(widget)

    def add_two_part_row(self, part1_label, part1_tooltip, part1_widget, part2_label, part2_tooltip, part2_widget):
        row_layout = QHBoxLayout()

        # Left part
        part1_layout = QHBoxLayout()
        self.add_tooltipped_widget_inline(part1_label, part1_tooltip, part1_widget, part1_layout)

        # Right part
        part2_layout = QHBoxLayout()
        self.add_tooltipped_widget_inline(part2_label, part2_tooltip, part2_widget, part2_layout)

        # Combine both parts in the row
        row_layout.addLayout(part1_layout)
        row_layout.addLayout(part2_layout)

        # Add the row layout to the main layout
        self.layout.addLayout(row_layout)

    def add_three_part_row(self, part1_label, part1_tooltip, part1_widget, part2_label, part2_tooltip, part2_widget,
                           part3_label, part3_tooltip, part3_widget):

        row_layout = QHBoxLayout()

        # Left part
        part1_layout = QHBoxLayout()
        self.add_tooltipped_widget_inline(part1_label, part1_tooltip, part1_widget, part1_layout)

        # Right part
        part2_layout = QHBoxLayout()
        self.add_tooltipped_widget_inline(part2_label, part2_tooltip, part2_widget, part2_layout)

        part3_layout = QHBoxLayout()
        self.add_tooltipped_widget_inline(part3_label, part3_tooltip, part3_widget, part3_layout)

        # Combine both parts in the row
        row_layout.addLayout(part1_layout)
        row_layout.addLayout(part2_layout)
        row_layout.addLayout(part3_layout)

        # Add the row layout to the main layout
        self.layout.addLayout(row_layout)

    def add_single_part_row(self, label_text, tooltip_text, widget):
        row_layout = QHBoxLayout()

        # Add label and widget inline
        label = QLabel(label_text)
        label.setToolTip(tooltip_text)
        row_layout.addWidget(label)
        row_layout.addWidget(widget)

        # Align all to the left
        row_layout.setAlignment(Qt.AlignLeft)

        # Add the row layout to the main layout
        self.layout.addLayout(row_layout)

    def create_csv_file_input(self):
        container = QWidget()
        layout = QHBoxLayout(container)
        line_edit = QLineEdit()
        button = QPushButton("Browse")
        button.clicked.connect(lambda: self.browse_csv_file(line_edit))
        layout.addWidget(line_edit)
        layout.addWidget(button)
        return container

    def create_folder_input(self):
        container = QWidget()
        layout = QHBoxLayout(container)
        line_edit = QLineEdit()
        button = QPushButton("Browse")
        button.clicked.connect(lambda: self.browse_folder(line_edit))
        layout.addWidget(line_edit)
        layout.addWidget(button)
        return container

    def create_double_spinbox(self, min_val, max_val, default_val=None):
        spinbox = QDoubleSpinBox()
        spinbox.setRange(min_val, max_val)
        if default_val is not None:
            spinbox.setValue(default_val)
        return spinbox

    def create_spinbox(self, min_val, max_val, default_val=0):
        spinbox = QSpinBox()
        spinbox.setRange(min_val, max_val)
        spinbox.setValue(default_val)
        return spinbox

    def create_combobox(self, items):
        combobox = QComboBox()
        combobox.addItems(items)
        return combobox

    def create_checkbox(self):
        return QCheckBox()

    def create_line_edit(self):
        return QLineEdit()

    def browse_csv_file(self, line_edit):
        file_path, _ = QFileDialog.getOpenFileName(self, "Select CSV File", filter="CSV Files (*.csv)")
        if file_path:
            line_edit.setText(file_path)

    def browse_folder(self, line_edit):
        folder_path = QFileDialog.getExistingDirectory(self, "Select Folder")
        if folder_path:
            line_edit.setText(folder_path)

    def build_command(self):
        """
        Constructs a command-line statement based on the current GUI input values.
        """
        command = f"python {os.path.abspath(__file__)}"

        # Helper function to get text or None if empty
        def get_text_or_none(widget):
            text = widget.findChild(QLineEdit).text()
            return text if text.strip() else None

        # Map GUI inputs to arguments
        arguments = {
            '--cp-output-csv': get_text_or_none(self.cp_output_csv_file),
            '--segmentation-results': get_text_or_none(self.segmentation_results_dir),
            '--neighbor-csv': get_text_or_none(self.neighbor_csv),
            '--interval-time': str(self.interval_time.value()) if self.interval_time.value() > 0 else None,
            '--doubling-time': str(
                self.doubling_time_of_bacteria.value()) if self.doubling_time_of_bacteria.value() > 0 else None,
            '--elongation-rate-method': self.elongation_rate_method.currentText(),
            '--pixel-per-micron': str(self.pixel_per_micron.value()),
            '--intensity-threshold': str(
                self.intensity_threshold.value()) if self.assigning_cell_type.isChecked() else None,
            '--assign-cell-type': False if not self.assigning_cell_type.isChecked() else True,
            '--disable-tracking-correction': False if not self.disable_tracking_correction.isChecked() else True,
            '--clf': self.clf.currentText(),
            '--num-cpus': str(self.n_cpu.value()),
            '--boundary-limits': self.boundary_limits.text().strip() if self.boundary_limits.text().strip() else None,
            '--dynamic-boundaries': get_text_or_none(self.dynamic_boundaries),
            '--output': get_text_or_none(self.out_dir),
            '--save_pickle': False if not self.save_pickle.isChecked() else True,
            '--verbose': False if not self.verbose.isChecked() else True,
        }

        # Build the command string
        for arg, value in arguments.items():
            if value is not None:
                if value == arg:  # For boolean flags (e.g., --verbose)
                    command += f" {value}"
                else:  # For key-value pairs
                    command += f" {arg} {value}"
        return command

    def run_analysis(self):
        # Helper function to get text or None if empty
        def get_text_or_none(widget):
            text = widget.findChild(QLineEdit).text()
            return text if text.strip() else None

        # Check if required inputs are provided
        missing_inputs = []

        # Check for essential inputs
        if not get_text_or_none(self.cp_output_csv_file):
            missing_inputs.append("CellProfiler Output CSV File")
        if not get_text_or_none(self.neighbor_csv):
            missing_inputs.append("Neighbor CSV File")
        if self.interval_time.value() <= 0:
            missing_inputs.append("Interval Time")
        if self.doubling_time_of_bacteria.value() <= 0:
            missing_inputs.append("Doubling Time")

        # Check segmentation directory if tracking correction is enabled
        if not self.disable_tracking_correction.isChecked() and not get_text_or_none(self.segmentation_results_dir):
            missing_inputs.append("Segmentation Results Folder (when tracking correction is enabled)")

        # Show warning if any required input is missing
        if missing_inputs:
            QMessageBox.warning(
                self,
                "Missing Inputs",
                f"The following required inputs are missing or invalid:\n- {', '.join(missing_inputs)}"
            )
            return

        # Construct the command statement
        command = self.build_command()

        # Run process_objects_data with validated inputs
        process_objects_data(
            cp_output_csv=get_text_or_none(self.cp_output_csv_file),
            segmentation_res_dir=get_text_or_none(self.segmentation_results_dir),
            neighbor_csv=get_text_or_none(self.neighbor_csv),
            interval_time=self.interval_time.value() if self.interval_time.value() > 0 else None,
            elongation_rate_method=self.elongation_rate_method.currentText(),
            pixel_per_micron=self.pixel_per_micron.value(),
            intensity_threshold=self.intensity_threshold.value() if self.assigning_cell_type.isChecked() else None,
            assigning_cell_type=self.assigning_cell_type.isChecked(),
            doubling_time=self.doubling_time_of_bacteria.value() if self.doubling_time_of_bacteria.value() > 0 else None,
            disable_tracking_correction=self.disable_tracking_correction.isChecked(),
            clf=self.clf.currentText(),
            n_cpu=self.n_cpu.value(),
            image_boundaries=self.boundary_limits.text() if self.boundary_limits.text().strip() else None,
            dynamic_boundaries=get_text_or_none(self.dynamic_boundaries),
            out_dir=get_text_or_none(self.out_dir),
            verbose=self.verbose.isChecked(),
            save_pickle=self.save_pickle.isChecked(),
            command=command
        )


def main():
    app = QApplication(sys.argv)
    gui = TrackRefinerGUI()
    gui.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



from scipy.optimize import linear_sum_assignment
import pandas as pd
import numpy as np
from Trackrefiner.core.correction.action.findOutlier import calculate_length_change_ratio_boundary, \
    calculate_lower_statistical_bound
from Trackrefiner.core.correction.action.helper import calculate_angles_between_slopes
from Trackrefiner.core.correction.modelTraning.calculation.iouCalForML import calculate_iou_ml
from Trackrefiner.core.correction.modelTraning.calculation.calcDistanceForML import calc_min_distance_ml
from Trackrefiner.core.correction.modelTraning.calculation.lengthRatio import check_len_ratio
from Trackrefiner.core.correction.modelTraning.calculation.calMotionAlignmentAngle import calc_motion_alignment_angle_ml
from Trackrefiner.core.correction.action.neighborAnalysis import compare_neighbor_sets
from Trackrefiner.core.correction.action.helper import calculate_trajectory_angles


def optimize_assignment_using_hungarian(cost_df):
    """
    Optimizes the assignment of rows to columns in a cost matrix using the Hungarian algorithm.

    :param pandas.DataFrame cost_df:
        A cost matrix represented as a DataFrame

    :returns:
        pandas.DataFrame

        A DataFrame containing the optimized assignments and their associated costs:

    """

    # Convert DataFrame to NumPy array
    cost_matrix = cost_df.values

    # Applying the Hungarian algorithm
    row_ind, col_ind = linear_sum_assignment(cost_matrix)

    # Retrieve the original row indices and column names
    original_row_indices = cost_df.index[row_ind]
    original_col_names = cost_df.columns[col_ind]

    # Retrieve the costs for the selected elements
    selected_costs = cost_matrix[row_ind, col_ind]

    # Create a DataFrame for the results
    result_df = pd.DataFrame({
        'idx1': original_row_indices,
        'idx2': original_col_names,
        'cost': selected_costs
    })

    return result_df


def calc_division_link_cost(df, neighbors_df, neighbor_list_array, df_source_daughter, center_coord_cols,
                            col_source, col_target, parent_image_number_col, parent_object_number_col,
                            division_links_model, division_vs_continuity_model, maintain_exist_link_cost_df,
                            check_maintenance_for='target', coordinate_array=None):

    """
    Calculates the cost of establishing division links between a source bacterium and its potential daughter bacteria.
    This function evaluates potential division links by calculating various spatial, geometric, and
    neighborhood-based features. It uses machine learning models to assess the likelihood of a division event
    based on these features.

    **Behavior**:
    - Calculates features like intersection-over-union (IoU), centroid distance, length ratios, and neighbor overlap.
    - Predicts division probabilities using a pre-trained model (`division_links_model`).
    - Compares division and continuity probabilities using a comparison model (`division_vs_continuity_model`).
    - Adjusts costs based on whether maintaining existing links is preferable.

    :param pandas.DataFrame df:
        Full DataFrame containing bacterial features for the current time step.
    :param pandas.DataFrame neighbors_df:
        DataFrame describing relationships between neighboring bacteria.
    :param lil_matrix neighbor_list_array:
        Sparse matrix representing neighbor relationships.
    :param pandas.DataFrame df_source_daughter:
        DataFrame containing the source bacteria and their candidate daughters.
    :param dict center_coord_cols:
        Dictionary specifying column names for bacterial centroid coordinates
        (e.g., `{"x": "Center_X", "y": "Center_Y"}`).
    :param str col_source:
        Suffix for source bacterium columns.
    :param str col_target:
        Suffix for target bacterium (daughter) columns.
    :param str parent_image_number_col:
        Column name for the parent image number.
    :param str parent_object_number_col:
        Column name for the parent object number.
    :param sklearn.Model division_links_model:
        Machine learning model to predict division probabilities.
    :param sklearn.Model division_vs_continuity_model:
        Machine learning model used to compare divided and non-divided states for bacteria.
    :param pandas.DataFrame maintain_exist_link_cost_df:
        Cost of maintaining current links between bacteria.
    :param str check_maintenance_for:
        Specifies whether maintenance costs are evaluated for the `target` or `source`. Default is `'target'`.
    :param csr_matrix coordinate_array:
        spatial coordinate matrix for neighborhood and overlap calculations.

    :returns:
        pandas.DataFrame:

        A cost matrix for potential division links, with rows as source bacteria and columns as their
        candidate daughters.
    """

    df_source_daughter[f'index_prev{col_target}'] = df_source_daughter[f'index{col_target}']
    df_source_daughter[f'index2{col_target}'] = df_source_daughter.index.values

    df_source_daughter[f'Length_Change_Ratio{col_target}'] = \
        (df_source_daughter[f'AreaShape_MajorAxisLength{col_target}'] /
         df_source_daughter[f'AreaShape_MajorAxisLength{col_source}'])

    df_source_daughter = calculate_iou_ml(df_source_daughter, col_source=f'prev_index{col_source}',
                                          col_target=f'prev_index{col_target}', link_type='div',
                                          coordinate_array=coordinate_array, both=True)

    df_source_daughter = calc_min_distance_ml(df_source_daughter, center_coord_cols, postfix_target=col_target,
                                              postfix_source=col_source, link_type='div')

    df_source_daughter[f'prev_time_step_index{col_target}'] = df_source_daughter[f'index{col_source}'].astype('int64')

    df_source_daughter[f'Neighbor_Difference_Count{col_target}'] = np.nan
    df_source_daughter[f'other_daughter_index{col_target}'] = np.nan
    df_source_daughter[f'parent_id{col_target}'] = df_source_daughter[f'id{col_source}']

    df_source_daughter = compare_neighbor_sets(df_source_daughter, neighbor_list_array, parent_image_number_col,
                                               parent_object_number_col, selected_rows_df=df_source_daughter,
                                               selected_time_step_df=df, return_common_elements=True,
                                               col_target=col_target)

    df_source_daughter[f'diff_common_diff{col_target}'] = \
        df_source_daughter[f'Neighbor_Difference_Count{col_target}'] - \
        df_source_daughter[f'Neighbor_Shared_Count{col_target}']

    df_source_daughter.loc[df_source_daughter[f'diff_common_diff{col_target}'] < 0, f'diff_common_diff{col_target}'] = 0

    df_source_daughter[f'angle_mother_daughter{col_target}'] = \
        calculate_angles_between_slopes(df_source_daughter[f'Bacterium_Slope{col_source}'].values,
                                        df_source_daughter[f'Bacterium_Slope{col_target}'].values)

    df_source_daughter[f'Prev_Center_X{col_target}'] = \
        df_source_daughter[f"{center_coord_cols['x']}{col_source}"]
    df_source_daughter[f'Prev_Center_Y{col_target}'] = \
        df_source_daughter[f"{center_coord_cols['y']}{col_source}"]

    direction_of_motion = calculate_trajectory_angles(df_source_daughter, center_coord_cols, suffix1=col_target)

    df_source_daughter[f'Direction_of_Motion{col_target}'] = direction_of_motion

    # motion alignment
    # calculated for original df and we should calc for new df
    df_source_daughter[f"Motion_Alignment_Angle{col_target}"] = np.nan
    df_source_daughter = calc_motion_alignment_angle_ml(df, neighbors_df, center_coord_cols,
                                                        selected_rows=df_source_daughter, col_target=col_target,
                                                        col_source=col_source)

    df_source_daughter['adjusted_Neighbor_Shared_Count' + col_target] = \
        df_source_daughter['Neighbor_Shared_Count' + col_target].replace(0, 1)

    df_source_daughter['neighbor_ratio' + col_target] = \
        (df_source_daughter['Neighbor_Difference_Count' + col_target] / (
            df_source_daughter['adjusted_Neighbor_Shared_Count' + col_target]))

    raw_features = ['iou', 'iou_continuity', 'min_distance', 'min_distance_continuity',
                    'Neighbor_Difference_Count' + col_target,
                    'neighbor_ratio' + col_target, 'Neighbor_Shared_Count' + col_target,
                    'angle_mother_daughter' + col_target, 'Direction_of_Motion' + col_target,
                    'Motion_Alignment_Angle' + col_target, 'Length_Change_Ratio' + col_target, 'index' + col_source,
                    'index_prev' + col_target, 'divideFlag' + col_target, 'LifeHistory' + col_target,
                    'age' + col_target, 'diff_common_diff' + col_target]

    df_source_daughter = df_source_daughter[raw_features].copy()

    df_source_daughter = df_source_daughter.rename(
        {
            'Neighbor_Difference_Count' + col_target: 'Neighbor_Difference_Count',
            'Neighbor_Shared_Count' + col_target: 'Neighbor_Shared_Count',
            'neighbor_ratio' + col_target: 'neighbor_ratio',
            'angle_mother_daughter' + col_target: 'angle_mother_daughter',
            'Direction_of_Motion' + col_target: 'Direction_of_Motion',
            'Motion_Alignment_Angle' + col_target: 'Motion_Alignment_Angle',
            'Length_Change_Ratio' + col_target: 'Length_Change_Ratio',
            'index_prev' + col_target: 'index_prev',
            'divideFlag' + col_target: 'divideFlag',
            'LifeHistory' + col_target: 'LifeHistory',
            'age' + col_target: 'age',
            'diff_common_diff' + col_target: 'diff_common_diff'
        }, axis=1)

    feature_list_division_links_model = ['iou', 'min_distance', 'neighbor_ratio', 'angle_mother_daughter']

    y_prob_division_links_model = \
        division_links_model.predict_proba(df_source_daughter[feature_list_division_links_model])[:, 1]
    df_source_daughter['prob_divided_bac_model'] = y_prob_division_links_model

    # difference_neighbors
    feature_list_for_compare = \
        ['iou', 'min_distance', 'neighbor_ratio', 'Direction_of_Motion', 'Motion_Alignment_Angle',
         'Length_Change_Ratio']

    # division is class 0
    y_prob_compare = \
        division_vs_continuity_model.predict_proba(df_source_daughter[feature_list_for_compare])[:, 0]
    df_source_daughter['prob_compare'] = y_prob_compare

    if division_vs_continuity_model is not None:

        df_source_daughter = df_source_daughter.loc[(df_source_daughter['prob_divided_bac_model'] > 0.5) &
                                                    (df_source_daughter['prob_compare'] > 0.5)]

        if df_source_daughter.shape[0] > 0:
            # Pivot this DataFrame to get the desired structure
            division_cost_df = \
                df_source_daughter[['index' + col_source, 'index_prev', 'prob_compare']].pivot(
                    index='index' + col_source, columns='index_prev', values='prob_compare')
            division_cost_df.columns.name = None
            division_cost_df.index.name = None

            if check_maintenance_for == 'target':
                # I want to measure is it possible to remove the current link of target bacterium?
                candidate_target_maintenance_cost = \
                    maintain_exist_link_cost_df.loc[:, maintain_exist_link_cost_df.columns.isin(
                        df_source_daughter['index_prev'].unique())]

                # don't need to convert probability to 1 - probability
                for col in candidate_target_maintenance_cost.columns.values:
                    non_na_probability = candidate_target_maintenance_cost[col].dropna().iloc[0]
                    division_cost_df.loc[division_cost_df[col] <= non_na_probability, col] = np.nan

            # for maintenance_to_be_check = source we can not check, and we should check it after

            division_cost_df = 1 - division_cost_df

        else:
            division_cost_df = pd.DataFrame()
    else:
        # Pivot this DataFrame to get the desired structure
        division_cost_df = \
            df_source_daughter[['index' + col_source, 'index_prev', 'prob_divided_bac_model']].pivot(
                index='index' + col_source, columns='index_prev', values='prob_divided_bac_model')
        division_cost_df.columns.name = None
        division_cost_df.index.name = None

    return division_cost_df


def calc_continuity_link_cost(df, neighbors_df, neighbor_list_array, source_bac_with_can_target,
                              center_coord_cols, col_source, col_target, parent_image_number_col,
                              parent_object_number_col, continuity_links_model, division_vs_continuity_model,
                              maintain_exist_link_cost_df, check_maintenance_for='target', coordinate_array=None):

    """
    Calculates the cost of establishing continuity link between a source bacterium and its potential target bacterium.
    This function evaluates continuity links by analyzing spatial, geometric, and neighborhood-based features.
    It determines the likelihood that a bacterium in one time step corresponds to the same bacterium in the next
    time step.

    **Behavior**:
    - Calculates features like intersection-over-union (IoU), centroid distance, length ratios, and neighbor overlap.
    - Predicts continuity probabilities using a pre-trained model (`continuity_links_model`).
    - Compares continuity and division probabilities using a comparison model (`division_vs_continuity_model`).
    - Adjusts costs based on whether maintaining existing links is preferable.

    :param pandas.DataFrame df:
        Full DataFrame containing bacterial features for the current time step.
    :param pandas.DataFrame neighbors_df:
        DataFrame describing relationships between neighboring bacteria.
    :param lil_matrix neighbor_list_array:
        Sparse matrix representing neighbor relationships.
    :param pandas.DataFrame source_bac_with_can_target:
        DataFrame containing the source bacteria and their candidate targets in the next time step.
    :param dict center_coord_cols:
        Dictionary specifying column names for bacterial centroid coordinates
        (e.g., `{"x": "Center_X", "y": "Center_Y"}`).
    :param str col_source:
        Suffix for source bacterium columns.
    :param str col_target:
        Suffix for target bacterium columns.
    :param str parent_image_number_col:
        Column name for the parent image number.
    :param str parent_object_number_col:
        Column name for the parent object number.
    :param sklearn.Model continuity_links_model:
        Machine learning model to predict continuity probabilities.
    :param sklearn.Model division_vs_continuity_model:
        Machine learning model used to compare divided and non-divided states for bacteria.
    :param pandas.DataFrame maintain_exist_link_cost_df:
        Cost of maintaining current links between bacteria.
    :param str check_maintenance_for:
        Specifies whether maintenance costs are evaluated for the `target` or `source`. Default is `'target'`.
    :param csr_matrix coordinate_array:
        spatial coordinate matrix for neighborhood and overlap calculations.

    :returns:
        pandas.DataFrame:

        A cost matrix for potential continuity links, with rows as source bacteria and columns as their
        candidate targets.
    """

    bac_len_to_bac_ratio_boundary = calculate_length_change_ratio_boundary(df)

    lower_bound_threshold = calculate_lower_statistical_bound(bac_len_to_bac_ratio_boundary)

    source_bac_with_can_target['index_prev' + col_target] = source_bac_with_can_target['index' + col_target]
    source_bac_with_can_target['index2' + col_target] = source_bac_with_can_target.index.values

    source_bac_with_can_target['Length_Change_Ratio' + col_target] = (
            source_bac_with_can_target['AreaShape_MajorAxisLength' + col_target] /
            source_bac_with_can_target['AreaShape_MajorAxisLength' + col_source])

    source_bac_with_can_target = \
        source_bac_with_can_target.loc[source_bac_with_can_target['Length_Change_Ratio' + col_target]
                                       >= lower_bound_threshold].copy()

    if source_bac_with_can_target.shape[0] > 0:

        source_bac_with_can_target = calculate_iou_ml(source_bac_with_can_target, col_source='prev_index' + col_source,
                                                      col_target='prev_index' + col_target, link_type='continuity',
                                                      coordinate_array=coordinate_array)

        source_bac_with_can_target = calc_min_distance_ml(source_bac_with_can_target, center_coord_cols,
                                                          postfix_target=col_target, postfix_source=col_source,
                                                          link_type=None)

        source_bac_with_can_target['prev_time_step_index' + col_target] = \
            source_bac_with_can_target['index' + col_source].astype('int64')

        source_bac_with_can_target['Neighbor_Difference_Count' + col_target] = np.nan
        source_bac_with_can_target['other_daughter_index' + col_target] = np.nan
        source_bac_with_can_target['parent_id' + col_target] = source_bac_with_can_target['id' + col_source]

        source_bac_with_can_target = \
            compare_neighbor_sets(source_bac_with_can_target, neighbor_list_array,
                                  parent_image_number_col, parent_object_number_col,
                                  selected_rows_df=source_bac_with_can_target, selected_time_step_df=df,
                                  return_common_elements=True, col_target=col_target)

        source_bac_with_can_target = check_len_ratio(df, source_bac_with_can_target, col_target=col_target,
                                                     col_source=col_source)

        # motion alignment
        # calculated for original df and we should calc for new df
        source_bac_with_can_target["Motion_Alignment_Angle" + col_target] = np.nan
        source_bac_with_can_target = \
            calc_motion_alignment_angle_ml(df, neighbors_df, center_coord_cols,
                                           selected_rows=source_bac_with_can_target, col_target=col_target,
                                           col_source=col_source)

        source_bac_with_can_target['Prev_Center_X' + col_target] = \
            source_bac_with_can_target[center_coord_cols['x'] + col_source]
        source_bac_with_can_target['Prev_Center_Y' + col_target] = \
            source_bac_with_can_target[center_coord_cols['y'] + col_source]

        direction_of_motion = calculate_trajectory_angles(source_bac_with_can_target,
                                                          center_coord_cols, suffix1=col_target)

        source_bac_with_can_target['Direction_of_Motion' + col_target] = direction_of_motion

        source_bac_with_can_target['adjusted_Neighbor_Shared_Count' + col_target] = \
            source_bac_with_can_target['Neighbor_Shared_Count' + col_target].replace(0, 1)

        source_bac_with_can_target['neighbor_ratio' + col_target] = \
            (source_bac_with_can_target['Neighbor_Difference_Count' + col_target] / (
                source_bac_with_can_target['adjusted_Neighbor_Shared_Count' + col_target]))

        raw_feature_list = ['iou', 'min_distance', 'Neighbor_Difference_Count' + col_target, 'Neighbor_Shared_Count' + col_target,
                            'neighbor_ratio' + col_target, 'length_dynamic' + col_target,
                            'Motion_Alignment_Angle' + col_target,
                            'Direction_of_Motion' + col_target, 'Length_Change_Ratio' + col_target,
                            'index' + col_source, 'index_prev' + col_target]

        source_bac_with_can_target = source_bac_with_can_target[raw_feature_list].copy()
        source_bac_with_can_target = source_bac_with_can_target.rename(
            {
                'Neighbor_Shared_Count' + col_target: 'Neighbor_Shared_Count',
                'Neighbor_Difference_Count' + col_target: 'Neighbor_Difference_Count',
                'neighbor_ratio' + col_target: 'neighbor_ratio',
                'length_dynamic' + col_target: 'length_dynamic',
                'Motion_Alignment_Angle' + col_target: 'Motion_Alignment_Angle',
                'Direction_of_Motion' + col_target: 'Direction_of_Motion',
                'Length_Change_Ratio' + col_target: 'Length_Change_Ratio',
                'index_prev' + col_target: 'index_prev',
            }, axis=1)

        # difference_neighbors
        feature_list_for_continuity_links_model = \
            ['iou', 'min_distance', 'neighbor_ratio', 'length_dynamic', 'Motion_Alignment_Angle']

        # difference_neighbors
        feature_list_for_compare = \
            ['iou', 'min_distance', 'neighbor_ratio', 'Direction_of_Motion', 'Motion_Alignment_Angle',
             'Length_Change_Ratio']

        y_prob_continuity_links_model = \
            continuity_links_model.predict_proba(source_bac_with_can_target[
                                                    feature_list_for_continuity_links_model])[:, 1]
        source_bac_with_can_target['prob_non_divided_bac_model'] = y_prob_continuity_links_model

        # non divided class 1
        y_prob_compare = \
            division_vs_continuity_model.predict_proba(source_bac_with_can_target[
                                                                  feature_list_for_compare])[:, 1]
        source_bac_with_can_target['prob_compare'] = y_prob_compare

        source_bac_with_can_target = \
            source_bac_with_can_target.loc[(source_bac_with_can_target['prob_non_divided_bac_model'] > 0.5) &
                                           (source_bac_with_can_target['prob_compare'] > 0.5) &
                                           (source_bac_with_can_target['Neighbor_Difference_Count'] <=
                                            source_bac_with_can_target['Neighbor_Shared_Count'])]

        # Pivot this DataFrame to get the desired structure
        continuity_link_cost_df = \
            source_bac_with_can_target[['index' + col_source, 'index_prev', 'prob_compare']].pivot(
                index='index' + col_source, columns='index_prev', values='prob_compare')
        continuity_link_cost_df.columns.name = None
        continuity_link_cost_df.index.name = None

        if check_maintenance_for == 'target':
            # I want to measure is it possible to remove the current link of target bacterium?
            candidate_target_maintenance_cost = \
                maintain_exist_link_cost_df.loc[:, maintain_exist_link_cost_df.columns.isin(
                    source_bac_with_can_target['index_prev'].unique())]

            # don't need to convert probability to 1 - probability
            for col in candidate_target_maintenance_cost.columns.values:
                this_target_bac_maintenance_cost = candidate_target_maintenance_cost[col].dropna().iloc[0]
                continuity_link_cost_df.loc[
                    continuity_link_cost_df[col] <= this_target_bac_maintenance_cost, col] = np.nan

        elif check_maintenance_for == 'source':

            # I want to measure is it possible to remove the current link of target bacterium?
            candidate_target_maintenance_cost = \
                maintain_exist_link_cost_df.loc[maintain_exist_link_cost_df.index.isin(
                    source_bac_with_can_target['index' + col_source].unique())]

            selected_candidate_target_maintenance_cost = \
                candidate_target_maintenance_cost[candidate_target_maintenance_cost.min(axis=1) > 0.5]

            for source_idx in selected_candidate_target_maintenance_cost.index.values:
                max_probability_value = candidate_target_maintenance_cost.loc[source_idx].max()

                continuity_link_cost_df.loc[
                    source_idx, continuity_link_cost_df.loc[source_idx] <= max_probability_value] = np.nan

        continuity_link_cost_df = 1 - continuity_link_cost_df

    else:
        continuity_link_cost_df = pd.DataFrame()

    return continuity_link_cost_df


def calc_division_link_cost_for_restoring_links(df, neighbors_df, neighbor_list_array, df_source_daughter,
                                                center_coord_cols, col_source, col_target, parent_image_number_col,
                                                parent_object_number_col, division_links_model, coordinate_array):

    """
    Calculates the cost of restoring division links between a source bacterium and its potential daughter
    bacteria (UB bacteria). This function evaluates potential division links by calculating various spatial,
    geometric, and neighborhood-based features.
    It uses machine learning model to assess the likelihood of a division event based on these features.

    **Behavior**:
    - Calculates features like intersection-over-union (IoU), centroid distance, length ratios, and neighbor overlap.
    - Predicts division probabilities using a pre-trained model (`division_links_model`).

    :param pandas.DataFrame df:
        Full DataFrame containing bacterial features for the current time step.
    :param pandas.DataFrame neighbors_df:
        DataFrame describing relationships between neighboring bacteria.
    :param lil_matrix neighbor_list_array:
        Sparse matrix representing neighbor relationships.
    :param pandas.DataFrame df_source_daughter:
        DataFrame containing the source bacteria and their candidate daughters.
    :param dict center_coord_cols:
        Dictionary specifying column names for bacterial centroid coordinates
        (e.g., `{"x": "Center_X", "y": "Center_Y"}`).
    :param str col_source:
        Suffix for source bacterium columns.
    :param str col_target:
        Suffix for target bacterium (daughter) columns.
    :param str parent_image_number_col:
        Column name for the parent image number.
    :param str parent_object_number_col:
        Column name for the parent object number.
    :param sklearn.Model division_links_model:
        Machine learning model to predict division probabilities.
    :param csr_matrix coordinate_array:
        spatial coordinate matrix for neighborhood and overlap calculations.

    :returns:
        pandas.DataFrame:

        A cost matrix for potential division links, with rows as source bacteria and columns as their
        candidate daughters.
    """

    df_source_daughter['index_prev' + col_target] = df_source_daughter['index' + col_target]
    df_source_daughter['index2' + col_target] = df_source_daughter.index.values

    df_source_daughter['Length_Change_Ratio' + col_target] = \
        (df_source_daughter['AreaShape_MajorAxisLength' + col_target] /
         df_source_daughter['AreaShape_MajorAxisLength' + col_source])

    df_source_daughter = calculate_iou_ml(df_source_daughter, col_source='prev_index' + col_source,
                                          col_target='prev_index' + col_target, link_type='div',
                                          coordinate_array=coordinate_array, both=True)

    df_source_daughter = calc_min_distance_ml(df_source_daughter, center_coord_cols, postfix_target=col_target,
                                              postfix_source=col_source, link_type='div')

    df_source_daughter['prev_time_step_index' + col_target] = \
        df_source_daughter['index' + col_source].fillna(-1).astype('int64')

    df_source_daughter['Neighbor_Difference_Count' + col_target] = np.nan
    df_source_daughter['other_daughter_index' + col_target] = np.nan
    df_source_daughter['parent_id' + col_target] = df_source_daughter['id' + col_source]

    df_source_daughter = \
        compare_neighbor_sets(df_source_daughter, neighbor_list_array,
                              parent_image_number_col, parent_object_number_col,
                              selected_rows_df=df_source_daughter, selected_time_step_df=df,
                              return_common_elements=True, col_target=col_target)

    df_source_daughter['diff_common_diff' + col_target] = \
        df_source_daughter['Neighbor_Difference_Count' + col_target] - \
        df_source_daughter['Neighbor_Shared_Count' + col_target]

    df_source_daughter.loc[df_source_daughter['diff_common_diff' + col_target] < 0, 'diff_common_diff' + col_target] = 0

    df_source_daughter['angle_mother_daughter' + col_target] = \
        calculate_angles_between_slopes(df_source_daughter['Bacterium_Slope' + col_source].values,
                                        df_source_daughter['Bacterium_Slope' + col_target].values)

    df_source_daughter['Prev_Center_X' + col_target] = \
        df_source_daughter[center_coord_cols['x'] + col_source]
    df_source_daughter['Prev_Center_Y' + col_target] = \
        df_source_daughter[center_coord_cols['y'] + col_source]

    direction_of_motion = calculate_trajectory_angles(df_source_daughter,
                                                      center_coord_cols, suffix1=col_target)

    df_source_daughter['Direction_of_Motion' + col_target] = direction_of_motion

    # motion alignment
    # calculated for original df and we should calc for new df
    df_source_daughter["Motion_Alignment_Angle" + col_target] = np.nan
    df_source_daughter = \
        calc_motion_alignment_angle_ml(df, neighbors_df, center_coord_cols,
                                       selected_rows=df_source_daughter, col_target=col_target,
                                       col_source=col_source)

    df_source_daughter['adjusted_Neighbor_Shared_Count' + col_target] = np.where(
        df_source_daughter['Neighbor_Shared_Count' + col_target] == 0,
        df_source_daughter['Neighbor_Shared_Count' + col_target] + 1,
        df_source_daughter['Neighbor_Shared_Count' + col_target]
    )

    df_source_daughter['neighbor_ratio' + col_target] = \
        (df_source_daughter['Neighbor_Difference_Count' + col_target] / (
            df_source_daughter['adjusted_Neighbor_Shared_Count' + col_target]))

    raw_features = ['iou', 'iou_continuity', 'min_distance', 'min_distance_continuity',
                    'Neighbor_Difference_Count' + col_target,
                    'Neighbor_Shared_Count' + col_target, 'neighbor_ratio' + col_target,
                    'angle_mother_daughter' + col_target, 'Direction_of_Motion' + col_target,
                    'Motion_Alignment_Angle' + col_target, 'Length_Change_Ratio' + col_target, 'index' + col_source,
                    'index_prev' + col_target, 'divideFlag' + col_target, 'LifeHistory' + col_target,
                    'age' + col_target, 'diff_common_diff' + col_target]

    df_source_daughter = df_source_daughter[raw_features].copy()

    df_source_daughter = df_source_daughter.rename(
        {
            'Neighbor_Difference_Count' + col_target: 'Neighbor_Difference_Count',
            'Neighbor_Shared_Count' + col_target: 'Neighbor_Shared_Count',
            'neighbor_ratio' + col_target: 'neighbor_ratio',
            'angle_mother_daughter' + col_target: 'angle_mother_daughter',
            'Direction_of_Motion' + col_target: 'Direction_of_Motion',
            'Motion_Alignment_Angle' + col_target: 'Motion_Alignment_Angle',
            'Length_Change_Ratio' + col_target: 'Length_Change_Ratio',
            'index_prev' + col_target: 'index_prev',
            'divideFlag' + col_target: 'divideFlag',
            'LifeHistory' + col_target: 'LifeHistory',
            'age' + col_target: 'age',
            'diff_common_diff' + col_target: 'diff_common_diff'
        }, axis=1)

    # 'Neighbor_Difference_Count'
    feature_list_division_links_model = ['iou', 'min_distance', 'neighbor_ratio', 'angle_mother_daughter']

    y_prob_division_links_model = \
        division_links_model.predict_proba(df_source_daughter[feature_list_division_links_model])[:, 1]
    df_source_daughter['prob_divided_bac_model'] = y_prob_division_links_model

    df_source_daughter = df_source_daughter.loc[df_source_daughter['prob_divided_bac_model'] > 0.5]

    if df_source_daughter.shape[0] > 0:
        # Pivot this DataFrame to get the desired structure
        division_cost_df = \
            df_source_daughter[['index' + col_source, 'index_prev', 'prob_divided_bac_model']].pivot(
                index='index' + col_source, columns='index_prev', values='prob_divided_bac_model')
        division_cost_df.columns.name = None
        division_cost_df.index.name = None

        division_cost_df = 1 - division_cost_df

    else:
        division_cost_df = pd.DataFrame()

    return division_cost_df


def calc_continuity_link_cost_for_restoring_links(df, neighbors_df, neighbor_list_array, source_bac_with_can_target,
                                                  center_coord_cols, col_source, col_target,
                                                  parent_image_number_col, parent_object_number_col,
                                                  continuity_links_model, coordinate_array=None):

    """
    Calculates the cost of restoring continuity link between a source bacterium and its potential UB target bacterium.
    This function evaluates continuity links by analyzing spatial, geometric, and neighborhood-based features.
    It determines the likelihood that a bacterium in one time step corresponds to the same bacterium in the next
    time step.

    **Behavior**:
    - Calculates features like intersection-over-union (IoU), centroid distance, length ratios, and neighbor overlap.
    - Predicts continuity probabilities using a pre-trained model (`continuity_links_model`).

    :param pandas.DataFrame df:
        Full DataFrame containing bacterial features for the current time step.
    :param pandas.DataFrame neighbors_df:
        DataFrame describing relationships between neighboring bacteria.
    :param lil_matrix neighbor_list_array:
        Sparse matrix representing neighbor relationships.
    :param pandas.DataFrame source_bac_with_can_target:
        DataFrame containing the source bacteria and their candidate UB target in the next time step.
    :param dict center_coord_cols:
        Dictionary specifying column names for bacterial centroid coordinates
        (e.g., `{"x": "Center_X", "y": "Center_Y"}`).
    :param str col_source:
        Suffix for source bacterium columns.
    :param str col_target:
        Suffix for target bacterium columns.
    :param str parent_image_number_col:
        Column name for the parent image number.
    :param str parent_object_number_col:
        Column name for the parent object number.
    :param sklearn.Model continuity_links_model:
        Machine learning model to predict continuity probabilities.`.
    :param csr_matrix coordinate_array:
        spatial coordinate matrix for neighborhood and overlap calculations.

    :returns:
        pandas.DataFrame:

        A cost matrix for potential continuity links, with rows as source bacteria and columns as their
        candidate targets (UB bacteria).
    """

    source_bac_with_can_target['index_prev' + col_target] = source_bac_with_can_target['index' + col_target]
    source_bac_with_can_target['index2' + col_target] = source_bac_with_can_target.index.values

    source_bac_with_can_target['Length_Change_Ratio' + col_target] = (
            source_bac_with_can_target['AreaShape_MajorAxisLength' + col_target] /
            source_bac_with_can_target['AreaShape_MajorAxisLength' + col_source])

    if source_bac_with_can_target.shape[0] > 0:

        source_bac_with_can_target = calculate_iou_ml(source_bac_with_can_target, col_source='prev_index' + col_source,
                                                      col_target='prev_index' + col_target, link_type='continuity',
                                                      coordinate_array=coordinate_array)

        source_bac_with_can_target = calc_min_distance_ml(source_bac_with_can_target, center_coord_cols,
                                                          postfix_target=col_target, postfix_source=col_source,
                                                          link_type=None)

        source_bac_with_can_target['prev_time_step_index' + col_target] = \
            source_bac_with_can_target['index' + col_source].fillna(-1).astype('int64')

        source_bac_with_can_target['Neighbor_Difference_Count' + col_target] = np.nan
        source_bac_with_can_target['other_daughter_index' + col_target] = np.nan
        source_bac_with_can_target['parent_id' + col_target] = source_bac_with_can_target['id' + col_source]

        source_bac_with_can_target = \
            compare_neighbor_sets(source_bac_with_can_target, neighbor_list_array,
                                  parent_image_number_col, parent_object_number_col,
                                  selected_rows_df=source_bac_with_can_target, selected_time_step_df=df,
                                  return_common_elements=True, col_target=col_target)

        source_bac_with_can_target = check_len_ratio(df, source_bac_with_can_target, col_target=col_target,
                                                     col_source=col_source)

        # motion alignment
        # calculated for original df and we should calc for new df
        source_bac_with_can_target["Motion_Alignment_Angle" + col_target] = np.nan
        source_bac_with_can_target = \
            calc_motion_alignment_angle_ml(df, neighbors_df, center_coord_cols,
                                           selected_rows=source_bac_with_can_target, col_target=col_target,
                                           col_source=col_source)

        source_bac_with_can_target['Prev_Center_X' + col_target] = \
            source_bac_with_can_target[center_coord_cols['x'] + col_source]
        source_bac_with_can_target['Prev_Center_Y' + col_target] = \
            source_bac_with_can_target[center_coord_cols['y'] + col_source]

        direction_of_motion = calculate_trajectory_angles(source_bac_with_can_target,
                                                          center_coord_cols, suffix1=col_target)

        source_bac_with_can_target['Direction_of_Motion' + col_target] = direction_of_motion

        source_bac_with_can_target['adjusted_Neighbor_Shared_Count' + col_target] = np.where(
            source_bac_with_can_target['Neighbor_Shared_Count' + col_target] == 0,
            source_bac_with_can_target['Neighbor_Shared_Count' + col_target] + 1,
            source_bac_with_can_target['Neighbor_Shared_Count' + col_target]
        )

        source_bac_with_can_target['neighbor_ratio' + col_target] = \
            (source_bac_with_can_target['Neighbor_Difference_Count' + col_target] / (
                source_bac_with_can_target['adjusted_Neighbor_Shared_Count' + col_target]))

        raw_feature_list = ['iou', 'min_distance', 'Neighbor_Difference_Count' + col_target, 'Neighbor_Shared_Count' + col_target,
                            'length_dynamic' + col_target,
                            'neighbor_ratio' + col_target,
                            'Motion_Alignment_Angle' + col_target,
                            'Direction_of_Motion' + col_target, 'Length_Change_Ratio' + col_target,
                            'index' + col_source, 'index_prev' + col_target]

        source_bac_with_can_target = source_bac_with_can_target[raw_feature_list].copy()
        source_bac_with_can_target = source_bac_with_can_target.rename(
            {
                'Neighbor_Shared_Count' + col_target: 'Neighbor_Shared_Count',
                'Neighbor_Difference_Count' + col_target: 'Neighbor_Difference_Count',
                'neighbor_ratio' + col_target: 'neighbor_ratio',
                'length_dynamic' + col_target: 'length_dynamic',
                'Motion_Alignment_Angle' + col_target: 'Motion_Alignment_Angle',
                'Direction_of_Motion' + col_target: 'Direction_of_Motion',
                'Length_Change_Ratio' + col_target: 'Length_Change_Ratio',
                'index_prev' + col_target: 'index_prev',
            }, axis=1)

        # difference_neighbors
        feature_list_for_continuity_links_model = \
            ['iou', 'min_distance', 'neighbor_ratio', 'length_dynamic', 'Motion_Alignment_Angle']

        y_prob_continuity_links_model = \
            continuity_links_model.predict_proba(source_bac_with_can_target[
                                                    feature_list_for_continuity_links_model])[:, 1]
        source_bac_with_can_target['prob_non_divided_bac_model'] = y_prob_continuity_links_model

        source_bac_with_can_target = \
            source_bac_with_can_target.loc[(source_bac_with_can_target['prob_non_divided_bac_model'] > 0.5) &
                                           (source_bac_with_can_target['Neighbor_Difference_Count'] <=
                                            source_bac_with_can_target['Neighbor_Shared_Count'])]

        # Pivot this DataFrame to get the desired structure
        continuity_link_cost_df = \
            source_bac_with_can_target[['index' + col_source, 'index_prev', 'prob_non_divided_bac_model']].pivot(
                index='index' + col_source, columns='index_prev', values='prob_non_divided_bac_model')
        continuity_link_cost_df.columns.name = None
        continuity_link_cost_df.index.name = None

        continuity_link_cost_df = 1 - continuity_link_cost_df

    else:
        continuity_link_cost_df = pd.DataFrame()

    return continuity_link_cost_df



import pandas as pd
import time
import pickle
import os
import glob
from Trackrefiner.core.processing.bacterialLifeHistoryAnalysis import process_bacterial_life_and_family
from Trackrefiner.core.correction.findFixTrackingErrors import find_fix_tracking_errors
from Trackrefiner.core.correction.action.helper import identify_important_columns, print_progress_bar
from Trackrefiner.core.correction.action.drawDistributionPlot import draw_feature_distribution
from Trackrefiner.core.correction.action.checkValidity import check_validity
import psutil
import threading


def monitor_system_usage(stats, interval=1, stop_event=None):
    process = psutil.Process()
    while not stop_event.is_set():
        cpu_usage = psutil.cpu_percent(interval=None)
        memory_info = process.memory_info()
        memory_usage = memory_info.rss / (1024 * 1024 * 1024)  # Convert memory to GB

        # Store CPU and memory usage in a shared list
        stats['cpu'].append(cpu_usage)
        stats['memory'].append(memory_usage)

        time.sleep(interval)  # Interval to report CPU and memory usage


class Dict2Class(object):

    def __init__(self, my_dict):
        for key in my_dict:
            setattr(self, key, my_dict[key])


def write_log_file(log_list, path):

    log_list = [v for v in log_list if v != '']

    # Open a file for writing
    with open(f'{path}log.txt', 'w') as file:
        # Write each element to the file
        for element in log_list:
            file.write(element + '\n')


def create_pickle_files(df, path, assigning_cell_type):

    lineage = {}
    time_steps = sorted(df['stepNum'].unique())
    for t in time_steps:
        lineage_data_frame = df.loc[(df["stepNum"] == t) & df["parent_id"]]
        lineage.update(dict(zip(lineage_data_frame["id"].values, lineage_data_frame["parent_id"].values)))
        data = {}
        data_frame_current_time_step = df.loc[df["stepNum"] == t]
        data["stepNum"] = t
        data["lineage"] = lineage
        # start index from 1
        data_frame_current_time_step.index = data_frame_current_time_step['id']

        # select important columns
        if assigning_cell_type:
            data_frame_current_time_step = data_frame_current_time_step[
                ['ObjectNumber', 'id', 'label', 'cellType', 'divideFlag', 'cellAge', 'growthRate', 'LifeHistory',
                 'startVol', 'targetVol', 'pos', 'time', 'radius', 'length', 'dir', 'ends', 'strainRate',
                 'strainRate_rolling']]
        else:
            data_frame_current_time_step = data_frame_current_time_step[
                ['ObjectNumber', 'id', 'label', 'divideFlag', 'cellAge', 'growthRate', 'LifeHistory', 'startVol',
                 'targetVol', 'pos', 'time', 'radius', 'length', 'dir', 'ends', 'strainRate', 'strainRate_rolling']]
        # convert to dictionary
        df_to_dict = data_frame_current_time_step.to_dict('index')
        # change dictionary to class
        for dictionary_key in df_to_dict.keys():
            df_to_dict[dictionary_key] = Dict2Class(df_to_dict[dictionary_key])
        data["cellStates"] = df_to_dict
        write_to_pickle_file(data, path, t)


def write_to_pickle_file(data, path, time_step):

    if not os.path.exists(path):
        os.mkdir(path)

    output_file = f"{path}Trackrefiner.step-{time_step:06}.pickle"

    with open(output_file, 'wb') as export:
        pickle.dump(data, export, protocol=-1)


def process_objects_data(cp_output_csv, segmentation_res_dir, neighbor_csv, interval_time, doubling_time,
                         elongation_rate_method, pixel_per_micron, assigning_cell_type, intensity_threshold,
                         disable_tracking_correction, clf, n_cpu, image_boundaries, dynamic_boundaries, out_dir,
                         save_pickle, verbose, command):

    """
    Processes CellProfiler output data, performs tracking correction, assigns cell types,
    and calculates additional features related to the life history of bacteria.


    :param str cp_output_csv:
        Path to the CellProfiler output file in CSV format, which contains measured bacterial features.
    :param str segmentation_res_dir:
        Path to the directory containing npy files (segmentation results exported from the CellProfiler pipeline).
    :param str neighbor_csv:
        Path to a CSV file containing information about the neighbors of each object.
    :param str out_dir:
        Path to the directory where processed output files will be saved.
    :param float interval_time:
        Time interval between consecutive images (in minutes).
    :param float doubling_time:
        Minimum life history duration of bacteria (in minutes) for analysis.
    :param str elongation_rate_method:
        Method for calculating the elongation rate. Options:
        - 'Average': Computes average growth rate.
        - 'Linear Regression': Estimates growth rate using linear regression.
    :param float pixel_per_micron:
        Conversion factor for pixels to microns (e.g., default value is 0.144).
    :param bool assigning_cell_type:
        If True, assigns cell types to objects based on intensity thresholds.
    :param float intensity_threshold:
        Threshold to classify objects into specific cell types based on intensity values.
    :param bool disable_tracking_correction:
        If True, disables tracking correction and performs calculations directly on the CellProfiler output data.
    :param str clf:
        Name of the classifier to be used for model training during tracking correction.
        Options: 'LogisticRegression', 'GaussianProcessClassifier', 'C-Support Vector Classifier'.
    :param int n_cpu:
        Number of CPU cores to be used for parallel computing.
        - `-1` indicates that all available CPUs will be utilized.
    :param str image_boundaries:
        Boundary limits defined for all time steps to filter out objects outside the image area.
    :param str dynamic_boundaries:
        Path to a CSV file specifying boundary limits for each time step. The file should contain columns:
        `Time Step`, `Lower X Limit`, `Upper X Limit`, `Lower Y Limit`, `Upper Y Limit`.
    :param bool save_pickle:
        If True, results are saved in `.pickle` format.
    :param bool verbose:
        If True, displays warnings and additional details during processing.
    :param str command:
        User's Command Statement

    :returns:
        None. Processed data and outputs are saved to the specified ``out_dir``.

    :rtype: None

    """

    log_list = []

    # Shared dictionary to store CPU and memory usage stats
    stats = {
        'cpu': [],  # List to store CPU usage over time
        'memory': []  # List to store memory usage over time
    }

    # Set up an event to stop the monitoring thread when execution is done
    stop_event = threading.Event()

    # Start the monitoring in a separate thread
    monitoring_thread = threading.Thread(target=monitor_system_usage, args=(stats, 1, stop_event))
    monitoring_thread.start()

    try:
        # Recording the start time
        start_time = time.time()
        start_time_str = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(start_time))
        start_time_log = f"Trackrefiner Processing started at: {start_time_str}\n"
        print(start_time_log)
        log_list.append(start_time_log)

        command_log = f"User's Command: {command}\n"
        print(command_log)
        log_list.append(command_log)

        check_validity(cp_output_csv, segmentation_res_dir, neighbor_csv, interval_time, doubling_time,
                       elongation_rate_method, pixel_per_micron, intensity_threshold,
                       disable_tracking_correction, clf, n_cpu, image_boundaries, dynamic_boundaries)

        # Initial call to print 0% progress
        print_progress_bar(0, prefix='Progress:', suffix='', length=50)

        cp_output_df = pd.read_csv(cp_output_csv)

        (center_coord_cols, all_rel_center_coord_cols, parent_image_number_col, parent_object_number_col,
         label_col) = identify_important_columns(cp_output_df)

        # Check if tracking correction is disabled; if so, skip neighbor file and segmentation files loading

        if not os.path.exists(neighbor_csv):
            raise FileNotFoundError(f"Neighbors CSV file not found: {neighbor_csv}")

        neighbors_df = pd.read_csv(neighbor_csv)
        neighbors_df = neighbors_df.loc[
            neighbors_df['Relationship'] == 'Neighbors'][['First Image Number', 'First Object Number',
                                                          'Second Image Number', 'Second Object Number']]

        if not disable_tracking_correction:

            if segmentation_res_dir is not None:
                sorted_seg_npy_files_list = sorted(glob.glob(segmentation_res_dir + '/*.npy'))
            else:
                sorted_seg_npy_files_list = []

            if len(sorted_seg_npy_files_list) == 0:
                raise FileNotFoundError(f"Segmentation files in npy format not found: {segmentation_res_dir}")
        else:
            sorted_seg_npy_files_list = []

        if ((len(sorted_seg_npy_files_list) > 0 and neighbors_df.shape[0] > 0) or
                (disable_tracking_correction and neighbors_df.shape[0] > 0)):

            if out_dir is not None:
                out_dir = f"{out_dir}/"
            else:
                # Create the directory if it does not exist
                out_dir = os.path.dirname(cp_output_csv)
                out_dir = f"{out_dir}/Trackrefiner/"
                os.makedirs(out_dir, exist_ok=True)

            (cp_output_df, find_fix_errors_log, logs_df, neighbors_df, cell_type_array) = \
                find_fix_tracking_errors(cp_output_df=cp_output_df, sorted_seg_npy_files_list=sorted_seg_npy_files_list,
                                         neighbors_df=neighbors_df, center_coord_cols=center_coord_cols,
                                         all_rel_center_coord_cols=all_rel_center_coord_cols,
                                         parent_image_number_col=parent_image_number_col,
                                         parent_object_number_col=parent_object_number_col, label_col=label_col,
                                         pixel_per_micron=pixel_per_micron, intensity_threshold=intensity_threshold,
                                         assigning_cell_type=assigning_cell_type, interval_time=interval_time,
                                         doubling_time=doubling_time,
                                         disable_tracking_correction=disable_tracking_correction,
                                         clf=clf, n_cpu=n_cpu, image_boundaries=image_boundaries,
                                         dynamic_boundaries=dynamic_boundaries, out_dir=out_dir, verbose=verbose)

            log_list.extend(find_fix_errors_log)

            # process the tracking data
            processed_df, processed_df_with_specific_cols = \
                process_bacterial_life_and_family(cp_output_df, interval_time, elongation_rate_method,
                                                  assigning_cell_type, cell_type_array, label_col, center_coord_cols)

            sel_cols = [col for col in processed_df.columns if col not in
                        ['noise_bac', 'mother_rpl', 'daughter_rpl', 'source_mcl', 'target_mcl', 'prev_index',
                         'checked', 'bad_division_flag', 'ovd_flag', 'bad_daughters_flag', 'pos', 'ends',
                         'prev_id', 'prev_parent_id', 'prev_label', 'age', 'parent_index', 'index',
                         'daughters_index', 'other_daughter_index', 'prev_time_step_index']]

            excluded_patterns = [
                'TrackObjects_Displacement', 'TrackObjects_DistanceTraveled',
                'TrackObjects_FinalAge', 'TrackObjects_IntegratedDistance',
                'TrackObjects_Lifetime', 'TrackObjects_Linearity',
                'TrackObjects_TrajectoryX', 'TrackObjects_TrajectoryY'
            ]

            sel_cols = [col for col in sel_cols if not any(pattern in col for pattern in excluded_patterns)]

            processed_df = processed_df[sel_cols]

            # Define the ordered list of columns based on the provided structure
            ordered_columns = [
                "id", "parent_id",
                "time", "cellAge", "LifeHistory",
                "TrajectoryX", "TrajectoryY",
                "Endpoint1_X", "Endpoint1_Y", "Endpoint2_X", "Endpoint2_Y",
                "Prev_Center_X", "Prev_Center_Y", "Prev_Endpoint1_X", "Prev_Endpoint1_Y",
                "Prev_Endpoint2_X", "Prev_Endpoint2_Y",
                "Bacterium_Slope", "Orientation_Angle_Between_Slopes",
                "Direction_of_Motion", "Motion_Alignment_Angle",
                "Source_Neighbor_Avg_TrajectoryX", "Source_Neighbor_Avg_TrajectoryY",
                "divideFlag", "Division_TimeStep", "Division_Family_Count",
                "Daughter_Mother_Length_Ratio", "Total_Daughter_Mother_Length_Ratio",
                "Max_Daughter_Mother_Length_Ratio", "Daughter_Avg_TrajectoryX", "Daughter_Avg_TrajectoryY",
                "Neighbor_Difference_Count", "Neighbor_Shared_Count",
                "radius", "dir", "Average_Length", "Length_Change_Ratio", "Avg_Length_Change_Ratio",
                "Unexpected_End", "Unexpected_Beginning",
                "Velocity", "Instant_Velocity", "Average_Instant_Velocity",
                "Elongation_Rate", "Instant_Elongation_Rate",
                "strainRate", "strainRate_rolling",
                "cellType", "startVol", "targetVol",
                "Prev_Bacterium_Slope", "Prev_MajorAxisLength", "Bacterium_Movement",
            ]

            # Add any remaining columns at the end to preserve all data
            remaining_columns = [col for col in processed_df.columns if col not in ordered_columns]
            final_columns = remaining_columns + ordered_columns

            # Reorder the dataframe
            processed_df = processed_df[final_columns]

            print_progress_bar(10, prefix='Progress:', suffix='', length=50)

            if save_pickle:
                create_pickle_files(processed_df_with_specific_cols, f'{out_dir}/pickle_files/',
                                    assigning_cell_type)

            cp_out_base_name = os.path.basename(cp_output_csv).split('.')[0]
            # Common prefix for all output paths
            prefix = f"{out_dir}/Trackrefiner.{cp_out_base_name}_{elongation_rate_method.replace(' ', '_')}"

            path = f"{prefix}_analysis"
            path_logs = f"{prefix}_logs"
            path_neighbors = f"{prefix}_neighbors"

            # write to csv
            processed_df.to_csv(f"{path}.csv", index=False)

            if not disable_tracking_correction:
                logs_df.to_csv(f'{path_logs}.csv', index=False)
                neighbors_df[['First Image Number', 'First Object Number',
                              'Second Image Number', 'Second Object Number']].to_csv(
                    f'{path_neighbors}.csv', index=False)

            output_log = f"Output Directory: {out_dir}"
            print(output_log)
            log_list.append(output_log)

            # draw plots
            features_dict = {'LifeHistory': ["Cell Cycle Duration (min)", 'Frequency', 'lifehistory_based'],
                             'startVol': ["Birth Length (um)", 'Frequency', 'lifehistory_based'],
                             'targetVol': ["Final Length (um)", 'Frequency', 'lifehistory_based'],
                             'Elongation_Rate': ["Elongation Rate (um/min)", 'Frequency', 'lifehistory_based'],
                             'Velocity': ["Velocity (um/min)", 'Frequency', 'lifehistory_based'],
                             'Average_Length': ["Average Length (um)", 'Frequency', 'lifehistory_based'],
                             'AreaShape_MajorAxisLength': ["Length (um)", 'Frequency', 'bacteria_based']}
            draw_feature_distribution(processed_df, features_dict, label_col, interval_time, doubling_time, out_dir)

        else:
            log = "The segmentation result folder or neighbor file or both are not available!"
            print(log)
            log_list.append(log)

    finally:
        # Stop the monitoring thread after all functions have executed
        stop_event.set()
        monitoring_thread.join()

        # Calculate maximum and average CPU and memory usage
        max_cpu = max(stats['cpu']) if stats['cpu'] else 0
        avg_cpu = sum(stats['cpu']) / len(stats['cpu']) if stats['cpu'] else 0
        max_memory = max(stats['memory']) if stats['memory'] else 0
        avg_memory = sum(stats['memory']) / len(stats['memory']) if stats['memory'] else 0

    end_time = time.time()
    end_time_str = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(end_time))

    end_time_log = f"Trackrefiner Process completed at: {end_time_str}"
    print(end_time_log)
    log_list.append(end_time_log)

    execution_time = end_time - start_time

    # Conversion to the respective time units
    days, remainder = divmod(execution_time, 86400)
    hours, remainder = divmod(remainder, 3600)
    minutes, seconds = divmod(remainder, 60)

    duration_log = f"Execution Time: {int(days)}:{int(hours)}:{int(minutes)}:{seconds} (day:hours:minutes:seconds)"
    print(duration_log)

    log_list.append(duration_log)

    resource_usage = "\nResource Usage Summary:"
    max_cpu_usage = f"Maximum CPU Usage: {max_cpu:.2f}%"
    avg_cpu_usage = f"Average CPU Usage: {avg_cpu:.2f}%"
    max_mem_usage = f"Maximum Memory Usage: {max_memory:.2f} GB"
    avg_mem_usage = f"Average Memory Usage: {avg_memory:.2f} GB"
    log_list.append(resource_usage)
    log_list.append(max_cpu_usage)
    log_list.append(avg_cpu_usage)
    log_list.append(max_mem_usage)
    log_list.append(avg_mem_usage)

    write_log_file(log_list, out_dir)
